"""
WEC Design Optimization Toolbox (*WecOptTool*) developed by
Sandia National Laboratories. See
`sandialabs.github.io/WecOptTool/ <https://sandialabs.github.io/WecOptTool/>`_.

The top-level :python:`wecopttool` module contains:

* The :python:`wecopttool.WEC` class, which is the main way to interact
  with *WecOptTool*.
* Support functions for basic functionality, accessed as
  :python:`wecoptool.<function>`.

Other functionalities are implemented in the submodules, and can be
accessed as :python:`wecopttool.<module>.<function>`.



**Type Aliases**

+-------------------------+----------------------------------------------------------------------------+
| Alias                   | Type                                                                       |
+=========================+============================================================================+
| :python:`StateFunction` | :python:`Callable[[WEC, np.ndarray, np.ndarray, xr.Dataset], np.ndarray]`  |
+-------------------------+----------------------------------------------------------------------------+
"""


from importlib.metadata import metadata as metadata
import logging
import warnings

from wecopttool.core import *
from wecopttool import waves
from wecopttool import hydrostatics
from wecopttool import pto

try:
  from wecopttool import geom
except ModuleNotFoundError:
  warnings.warn("`geom` submodule not loaded because of missing dependencies. Install these by running `pip install wecopttool[geometry]`.")


# metadata
_metadata = metadata('wecopttool')
__version__ = _metadata['Version']
__version_info__ = tuple(__version__.split('.'))
__title__ = _metadata['Name']
__description__ = _metadata['Summary']
__author__ = _metadata['Author']
__uri__ = _metadata['Project-URL'].split(',')[1].strip()
__license__ = _metadata['License']

# logging
_handler = logging.StreamHandler()
_formatter = logging.Formatter(logging.BASIC_FORMAT)
_handler.setFormatter(_formatter)

_log = logging.getLogger(__name__)
_log.addHandler(_handler)

_log_capytaine = logging.getLogger("capytaine")
_log_capytaine.addHandler(_handler)


def set_loglevel(
    level: str,
    wecopttool: bool = True,
    capytaine: bool = True,
) -> None:
    """ Change the logging level of the :python:`wecopttool` and
    :python:`capytaine` loggers to the specified level.

    Parameters
    ----------
    level
        Level for :py:meth:`python.logging.Logger.setLevel`.
        See `list of logging levels <https://docs.python.org/3/library/logging.html#levels>`_.
    wecopttool
        Whether to change WecOptTool's log level.
    capytaine
        Whether to change Capytaine's log level.
    """
    if wecopttool:
        _log.setLevel(level.upper())
    if capytaine:
        _log_capytaine.setLevel(level.upper())



"""Provide power take-off (PTO) forces and produced energy functions
for common PTO control approaches.

The PTO produced energy can be used as the objective function for the
control optimization.
The PTO force can be included as an additional force in the WEC
dynamics.

Contains:

* The *PTO* class
* Controller functions

"""


from __future__ import annotations


__all__ = [
    "PTO",
    "controller_unstructured",
    "controller_pid",
    "controller_pi",
    "controller_p",
]


from typing import Optional, TypeVar, Callable, Union

import autograd.numpy as np
from autograd.builtins import isinstance, tuple, list, dict
from autograd.numpy import ndarray
from scipy.linalg import block_diag
from scipy.optimize import OptimizeResult
from xarray import DataArray, Dataset
from datetime import datetime
from scipy.optimize import OptimizeResult

from wecopttool.core import complex_to_real, td_to_fd
from wecopttool.core import dofmat_to_vec, vec_to_dofmat
from wecopttool.core import TWEC, TStateFunction, FloatOrArray


# type aliases
TPTO = TypeVar("TPTO", bound="PTO")
TLOSS = Callable[[FloatOrArray, FloatOrArray], FloatOrArray]


class PTO:
    """A power take-off (PTO) object to be used in conjunction with a
    :py:class:`wecopttool.WEC` object.
    """

    def __init__(self,
        ndof: int,
        kinematics: Union[TStateFunction, ndarray],
        controller: Optional[TStateFunction] = None,
        impedance: Optional[ndarray] = None,
        loss: Optional[TLOSS] = None,
        names: Optional[list[str]] = None,
    ) -> None:
        """Create a PTO object.

        The :py:class:`wecopttool.pto.PTO` class describes the
        kinematics, control logic, impedance and/or non-linear power
        loss of a power take-off system.
        The forces/moments applied by a
        :py:class:`wecopttool.pto.PTO` object can be applied to a
        :py:class:`wecopttool.WEC` object through the
        :py:attr:`wecopttool.WEC.f_add` property.
        The power produced by a :py:class:`wecopttool.pto.PTO` object
        can be used for the :python:`obj_fun` of pseudo-spectral
        optimization problem when calling
        :py:meth:`wecopttool.WEC.solve`.

        Parameters
        ----------
        ndof
            Number of degrees of freedom.
        kinematics
            Transforms state from WEC to PTO frame. May be a matrix
            (for linear kinematics) or function (for nonlinear
            kinematics).
        controller
            Function with signature
            :python:`def fun(wec, x_wec, x_opt, waves, nsubsteps):`
            or matrix with shape (PTO DOFs, WEC DOFs) that converts
            from the WEC DOFs to the PTO DOFs.
        impedance
            Matrix representing the PTO impedance.
        loss
            Function that maps flow and effort variables to a
            non-linear power loss.
            The output is the dissipated power (loss) in Watts.
            This should be a positive value.
        names
            PTO names.
        """
        self._ndof = ndof
        # names
        if names is None:
            names = [f'PTO_{i}' for i in range(ndof)]
        elif ndof == 1 and isinstance(names, str):
            names = [names]
        self._names = names
        # kinematics
        if callable(kinematics):
            def kinematics_fun(wec, x_wec, x_opt, waves, nsubsteps=1):
                pos_wec = wec.vec_to_dofmat(x_wec)
                tmat = self._tmat(wec, nsubsteps)
                pos_wec_td = np.dot(tmat, pos_wec)
                return kinematics(pos_wec_td)
        else:
            def kinematics_fun(wec, x_wec, x_opt, waves, nsubsteps=1):
                n = wec.nt*nsubsteps
                return np.repeat(kinematics[:, :, np.newaxis], n, axis=-1)
        self._kinematics = kinematics_fun
        # controller
        if controller is None:
            controller = controller_unstructured

        def force(wec, x_wec, x_opt, waves, nsubsteps=1):
            return controller(self, wec, x_wec, x_opt, waves, nsubsteps)

        self._force = force

        # power
        if impedance is not None:
            check_1 = impedance.shape[0] == impedance.shape[1] == 2*self.ndof
            check_2 = len(impedance.shape) == 3
            if not (check_1 and check_2):
                raise TypeError(
                    "Impedance should have size [2*ndof, 2*ndof, nfreq]"
                )
            for i in range(impedance.shape[2]-1):
                check_3 = (
                    np.allclose(np.real(impedance[:, :, i+1]), np.real(impedance[:, :, 0]))
                )
                if not check_3:
                    raise ValueError(
                        "Real component of impedance must be constant for " +
                        "all frequencies."
                    )
            impedance_abcd = _make_abcd(impedance, ndof)
            self._transfer_mat = _make_mimo_transfer_mat(impedance_abcd, ndof)
        else:
            self._transfer_mat = None
        self._impedance = impedance
        self._loss = loss

    @property
    def ndof(self) -> int:
        """Number of degrees of freedom."""
        return self._ndof

    @property
    def names(self) -> ndarray:
        """DOF Names."""
        return self._names

    @property
    def kinematics(self) -> TStateFunction:
        """Kinematics function.
        """
        return self._kinematics

    @property
    def force(self) -> TStateFunction:
        """PTO force in PTO coordinates."""
        return self._force

    @property
    def impedance(self) -> ndarray:
        """Impedance matrix."""
        return self._impedance

    @property
    def loss(self) -> TLOSS:
        """Nonlinear power loss function with outputs in Watts."""
        return self._loss

    @property
    def transfer_mat(self) -> ndarray:
        """Transfer matrix."""
        return self._transfer_mat

    def _tmat(self, wec, nsubsteps: Optional[int] = 1):
        if nsubsteps==1:
            tmat = wec.time_mat
        else:
            tmat = wec.time_mat_nsubsteps(nsubsteps)
        return tmat

    def _fkinematics(self,
        f_wec: ndarray,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: Optional[ndarray] = None,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> ndarray:
        """Return time-domain values in the PTO frame.

        Parameters
        ----------
        f_wec
            Fourier coefficients of some quantity "f" in the WEC frame.
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        time_mat = self._tmat(wec, nsubsteps)
        f_wec_td = np.dot(time_mat, f_wec)
        assert f_wec_td.shape == (wec.nt*nsubsteps, wec.ndof)
        f_wec_td = np.expand_dims(np.transpose(f_wec_td), axis=0)
        kinematics_mat = self.kinematics(wec, x_wec, x_opt, waves, nsubsteps)
        return np.transpose(np.sum(kinematics_mat*f_wec_td, axis=1))

    def position(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> ndarray:
        """Calculate the PTO position time-series.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        pos_wec = wec.vec_to_dofmat(x_wec)
        return self._fkinematics(pos_wec, wec, x_wec, x_opt, waves, nsubsteps)

    def velocity(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> ndarray:
        """Calculate the PTO velocity time-series.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        pos_wec = wec.vec_to_dofmat(x_wec)
        vel_wec = np.dot(wec.derivative_mat, pos_wec)
        return self._fkinematics(vel_wec, wec, x_wec, x_opt, waves, nsubsteps)

    def acceleration(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> np.ndarray:
        """Calculate the PTO acceleration time-series.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        pos_wec = wec.vec_to_dofmat(x_wec)
        acc_wec = np.dot(wec.derivative2_mat, pos_wec)
        return self._fkinematics(acc_wec, wec, x_wec, x_opt, waves, nsubsteps)

    def force_on_wec(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> ndarray:
        """Calculate the PTO force on WEC.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        force_td = self.force(wec, x_wec, x_opt, waves, nsubsteps)
        assert force_td.shape == (wec.nt*nsubsteps, self.ndof)
        force_td = np.expand_dims(np.transpose(force_td), axis=0)
        assert force_td.shape == (1, self.ndof, wec.nt*nsubsteps)
        kinematics_mat = self.kinematics(wec, x_wec, x_opt, waves, nsubsteps)
        kinematics_mat = np.transpose(kinematics_mat, (1,0,2))
        return np.transpose(np.sum(kinematics_mat*force_td, axis=1))

    def mechanical_power(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> np.ndarray:
        """Calculate the mechanical power time-series in each PTO DOF
        for a given system state.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        force_td = self.force(wec, x_wec, x_opt, waves, nsubsteps)
        vel_td = self.velocity(wec, x_wec, x_opt, waves, nsubsteps)
        return vel_td * force_td

    def mechanical_energy(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> float:
        """Calculate the mechanical energy in each PTO DOF for a given
        system state.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        power_td = self.mechanical_power(wec, x_wec, x_opt, waves, nsubsteps)
        return np.sum(power_td) * wec.dt/nsubsteps

    def mechanical_average_power(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> float:
        """Calculate average mechanical power in each PTO DOF for a
        given system state.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        energy = self.mechanical_energy(wec, x_wec, x_opt, waves, nsubsteps)
        return energy / wec.tf

    def power_variables(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> tuple[ndarray, ndarray]:
        """Calculate the power variables (flow q and effort e) time-series
        in each PTO DOF for a given system state.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        # convert q1 (PTO velocity), e1 (PTO force)
        # to q2 (flow variable), e2 (effort variable)
        if self.impedance is not None:
            q1_td = self.velocity(wec, x_wec, x_opt, waves)
            e1_td = self.force(wec, x_wec, x_opt, waves)
            q1 = complex_to_real(td_to_fd(q1_td, False))
            e1 = complex_to_real(td_to_fd(e1_td, False))
            vars_1 = np.hstack([q1, e1])
            vars_1_flat = dofmat_to_vec(vars_1)
            vars_2_flat = np.dot(self.transfer_mat, vars_1_flat)
            vars_2 = vec_to_dofmat(vars_2_flat, 2*self.ndof)
            q2 = vars_2[:, :self.ndof]
            e2 = vars_2[:, self.ndof:]
            time_mat = self._tmat(wec, nsubsteps)
            q2_td = np.dot(time_mat, q2)
            e2_td = np.dot(time_mat, e2)
        else:
            q2_td = self.velocity(wec, x_wec, x_opt, waves, nsubsteps)
            e2_td = self.force(wec, x_wec, x_opt, waves, nsubsteps)
        return q2_td, e2_td

    def power(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> ndarray:
        """Calculate the power time-series in each PTO DOF for a given
        system state.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        q2_td, e2_td = self.power_variables(wec, x_wec,
                                            x_opt, waves, nsubsteps)
        # power
        power_out = q2_td * e2_td
        if self.loss is not None:
            power_out = power_out + self.loss(q2_td, e2_td)
        return power_out

    def energy(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> float:
        """Calculate the energy in each PTO DOF for a given system
        state.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        power_td = self.power(wec, x_wec, x_opt, waves, nsubsteps)
        return np.sum(power_td) * wec.dt/nsubsteps

    def average_power(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> float:
        """Calculate the average power in each PTO DOF for a given
        system state.

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        energy = self.energy(wec, x_wec, x_opt, waves, nsubsteps)
        return energy / wec.tf

    def transduced_flow(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> float:
        """Calculate the transduced flow variable time-series in each PTO DOF
        for a given system state. Equals the PTO velocity if no impedance
        is defined.

        Examples for PTO impedance and corresponding flow variables:

        - OWC: (pneumatic admittance)^-1 : flow = volumetric air flow

        - Drive-train: rotational impedance : flow = rotational velocity

        - Generator: winding impedance: flow = electric current

        - Drive-train and Generator combined: flow = electric current

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        q2_td, _ = self.power_variables(wec, x_wec,
                                        x_opt, waves, nsubsteps)
        return q2_td

    def transduced_effort(self,
        wec: TWEC,
        x_wec: ndarray,
        x_opt: ndarray,
        waves: Optional[Dataset] = None,
        nsubsteps: Optional[int] = 1,
    ) -> float:
        """Calculate the transduced flow variable time-series in each PTO DOF
        for a given system state. Equals the PTO force if no impedance
        is defined.

        Examples for PTO impedance and corresponding effort variables:

        - OWC: (pneumatic admittance)^-1 : effort =  air pressure

        - Drive-train: rotational impedance : effort = torque

        - Generator: winding impedance: effort = voltage

        - Drive-train and Generator combined: effort = voltage

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        x_wec
            WEC dynamic state.
        x_opt
            Optimization (control) state.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
        """
        _, e2_td = self.power_variables(wec, x_wec, x_opt, waves, nsubsteps)
        return e2_td

    def post_process(self,
        wec: TWEC,
        res: OptimizeResult,
        waves: Optional[DataArray] = None,
        nsubsteps: Optional[int] = 1,
    ) -> tuple[Dataset, Dataset]:
        """Transform the results from optimization solution to a form
        that the user can work with directly.

        Examples
        --------
        The :py:meth:`wecopttool.WEC.solve` method only returns the
        raw results dictionary produced by :py:func:`scipy.optimize.minimize`.

        >>> res_opt = wec.solve(waves=wave,
                                obj_fun=pto.average_power,
                                nstate_opt=2*nfreq+1)

        To get the post-processed results for the
        :py:class:`wecopttool.pto.PTO`, you may call

        >>> res_wec_fd, res_wec_td = wec.post_process(wec,res_opt)
        >>> res_pto_fd, res_pto_td = pto.post_process(wec,res_opt)

        For smoother plots, you can set :python:`nsubsteps` to a value
        greater than 1.

        >>> res_pto_fd, res_pto_td = pto.post_process(wec,res_opt,
                                                      nsubsteps=4)
        >>> res_pto_td.power.plot()

        Parameters
        ----------
        wec
            :py:class:`wecopttool.WEC` object.
        res
            Results produced by :py:func:`scipy.optimize.minimize`.
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.

        Returns
        -------
        results_fd
            :py:class:`xarray.Dataset` with frequency domain results.
        results_td
            :py:class:`xarray.Dataset` with time domain results.
        """
        create_time = f"{datetime.utcnow()}"

        x_wec, x_opt = wec.decompose_state(res.x)

        # position
        pos_td = self.position(wec, x_wec, x_opt, waves, nsubsteps)
        pos_fd = wec.td_to_fd(pos_td[::nsubsteps])

        # velocity
        vel_td = self.velocity(wec, x_wec, x_opt, waves, nsubsteps)
        vel_fd = wec.td_to_fd(vel_td[::nsubsteps])

        # acceleration
        acc_td = self.acceleration(wec, x_wec, x_opt, waves, nsubsteps)
        acc_fd = wec.td_to_fd(acc_td[::nsubsteps])

        # force
        force_td = self.force(wec, x_wec, x_opt, waves, nsubsteps)
        force_fd = wec.td_to_fd(force_td[::nsubsteps])

        # power
        elec_power_td = self.power(wec, x_wec, x_opt, waves, nsubsteps)
        elec_power_fd = wec.td_to_fd(elec_power_td[::nsubsteps])

        # mechanical power
        mech_power_td = self.mechanical_power(wec, x_wec, x_opt, waves,
                                              nsubsteps)
        mech_power_fd = wec.td_to_fd(mech_power_td[::nsubsteps])

        # stack mechanical and electrical power
        power_names = ['mech','elec']
        power_fd = np.stack((mech_power_fd,elec_power_fd))
        power_td = np.stack((mech_power_td,elec_power_td))

        pos_attr = {'long_name': 'Position', 'units': 'm or rad'}
        vel_attr = {'long_name': 'Velocity', 'units': 'm/s or rad/s'}
        acc_attr = {'long_name': 'Acceleration',
                    'units': 'm/s^2 or rad/s^2'}
        force_attr = {'long_name': 'Force or moment on WEC',
                      'units': 'N or Nm'}
        power_attr = {'long_name': 'Power', 'units': 'W'}
        mech_power_attr = {'long_name': 'Mechanical power', 'units': 'W'}
        omega_attr = {'long_name': 'Radial frequency', 'units': 'rad/s'}
        freq_attr = {'long_name': 'Frequency', 'units': 'Hz'}
        period_attr = {'long_name': 'Period', 'units': 's'}
        dof_attr = {'long_name': 'PTO degree of freedom'}
        time_attr = {'long_name': 'Time', 'units': 's'}
        type_attr = {'long_name': 'Power type'}

        t_dat = wec.time_nsubsteps(nsubsteps)

        results_fd = Dataset(
            data_vars={
                'pos': (['omega','dof'], pos_fd, pos_attr),
                'vel': (['omega','dof'], vel_fd, vel_attr),
                'acc': (['omega','dof'], acc_fd, acc_attr),
                'force': (['omega','dof'], force_fd, force_attr),
                'power': (['type','omega','dof'], power_fd, power_attr),
            },
            coords={
                'omega':('omega', wec.omega, omega_attr),
                'freq':('omega', wec.frequency, freq_attr),
                'period':('omega', wec.period, period_attr),
                'dof':('dof', self.names, dof_attr),
                'type':('type', power_names, power_attr)},
            attrs={"time_created_utc": create_time}
            )

        results_td = Dataset(
            data_vars={
                'pos': (['time','dof'], pos_td, pos_attr),
                'vel': (['time','dof'], vel_td, vel_attr),
                'acc': (['time','dof'], acc_td, acc_attr),
                'force': (['time','dof'], force_td, force_attr),
                'power': (['type','time','dof'], power_td, power_attr),
            },
            coords={
                'time':('time', t_dat, time_attr),
                'dof':('dof', self.names, dof_attr),
                'type':('type', power_names, power_attr)},
            attrs={"time_created_utc": create_time}
            )

        if self.impedance is not None:
        #transduced flow and effort variables
            q2_td, e2_td = self.power_variables(wec, x_wec, x_opt,
                                                waves, nsubsteps)
            q2_fd = wec.td_to_fd(q2_td[::nsubsteps])
            e2_fd = wec.td_to_fd(e2_td[::nsubsteps])

            q2_attr = {'long_name': 'Transduced Flow',
                       'units': 'A or m^3/s or rad/s or m/s'}
            e2_attr = {'long_name': 'Transduced Effort',
                       'units': 'V or N/m^2 or Nm or Ns'}

            results_td = results_td.assign({
                        'trans_flo': (['time','dof'], q2_td, q2_attr),
                        'trans_eff': (['time','dof'], e2_td, e2_attr),
                    })
            results_fd = results_fd.assign({
                        'trans_flo': (['omega','dof'], q2_fd, q2_attr),
                        'trans_eff': (['omega','dof'], e2_fd, e2_attr),
                    })


        return results_fd, results_td


# power conversion chain
def _make_abcd(impedance: ndarray, ndof: int) -> ndarray:
    """Transform the impedance matrix into ABCD form from a MIMO
    transfer function.

    Parameters
    ----------
    impedance
        Matrix representing the PTO impedance.
        Size 2*n_dof.
    ndof
        Number of degrees of freedom.
    """
    z_11 = impedance[:ndof, :ndof, :]  # Fu
    z_12 = impedance[:ndof, ndof:, :]  # Fi
    z_21 = impedance[ndof:, :ndof, :]  # Vu
    z_22 = impedance[ndof:, ndof:, :]  # Vi
    z_12_inv = np.linalg.inv(z_12.T).T

    mmult = lambda a,b: np.einsum('mnr,mnr->mnr', a, b)
    abcd_11 = -1 * mmult(z_12_inv, z_11)
    abcd_12 = z_12_inv
    abcd_21 = z_21 - mmult(z_22, mmult(z_12_inv, z_11))
    abcd_22 = mmult(z_22, z_12_inv)

    row_1 = np.hstack([abcd_11, abcd_12])
    row_2 = np.hstack([abcd_21, abcd_22])
    return np.vstack([row_1, row_2])


def _make_mimo_transfer_mat(
    impedance_abcd: ndarray,
    ndof: int,
) -> np.ndarray:
    """Create a block matrix of a MIMO transfer function.

    Parameters
    ----------
    impedance
        PTO impedance in ABCD form.
    ndof
        Number of degrees of freedom.
    """
    def block(re, im): return np.array([[re, -im], [im, re]])
    for idof in range(2*ndof):
        for jdof in range(2*ndof):
            Zp = impedance_abcd[idof, jdof, :]
            re = np.real(Zp)
            im = np.imag(Zp)
            # Exclude the sine component of the 2-point wave
            blocks = [block(ire, iim) for (ire, iim) in zip(re[:-1], im[:-1])]
            # re[0] added for the zero frequency power loss (DC), could be re[n]
            blocks = [re[0]] + blocks + [re[-1]]
            if jdof==0:
                row = block_diag(*blocks)
            else:
                row = np.hstack([row, block_diag(*blocks)])
        if idof==0:
            mat = row
        else:
            mat = np.vstack([mat, row])
    return mat


# controllers
def controller_unstructured(
    pto: TPTO,
    wec: TWEC,
    x_wec: ndarray,
    x_opt: ndarray,
    waves: Optional[Dataset] = None,
    nsubsteps: Optional[int] = 1,
) -> ndarray:
    """Unstructured numerical optimal controller that returns a time
    history of PTO forces.

    Parameters
    ----------
    pto
        :py:class:`wecopttool.pto.PTO` object.
    wec
        :py:class:`wecopttool.WEC` object.
    x_wec
        WEC dynamic state.
    x_opt
        Optimization (control) state.
    waves
        :py:class:`xarray.Dataset` with the structure and elements
        shown by :py:mod:`wecopttool.waves`.
    nsubsteps
        Number of steps between the default (implied) time steps.
        A value of :python:`1` corresponds to the default step
        length.
    """
    tmat = pto._tmat(wec, nsubsteps)
    x_opt = np.reshape(x_opt[:len(tmat[0])*pto.ndof], (-1, pto.ndof), order='F')
    return np.dot(tmat, x_opt)


def controller_pid(
    pto: TPTO,
    wec: TWEC,
    x_wec: ndarray,
    x_opt: ndarray,
    waves: Optional[Dataset] = None,
    nsubsteps: Optional[int] = 1,
    proportional: Optional[bool] = True,
    integral: Optional[bool] = True,
    derivative: Optional[bool] = True,
) -> ndarray:
    """Proportional-integral-derivative (PID) controller that returns
    a time history of PTO forces.

    Parameters
    ----------
    pto
        :py:class:`wecopttool.pto.PTO` object.
    wec
        :py:class:`wecopttool.WEC` object.
    x_wec
        WEC dynamic state.
    x_opt
        Optimization (control) state.
    waves
        :py:class:`xarray.Dataset` with the structure and elements shown
        by :py:mod:`wecopttool.waves`.
    nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
    proportional
        True to include proportional gain
    integral
        True to include integral gain
    derivative
        True to include derivative gain
    """
    ndof = pto.ndof
    force_td = np.zeros([wec.nt*nsubsteps, ndof])
    idx = 0

    def update_force_td(response):
        nonlocal idx, force_td
        gain = np.reshape(x_opt[idx*ndof:(idx+1)*ndof], [1, ndof])
        force_td = force_td + gain*response
        idx = idx + 1

    if proportional:
        vel_td = pto.velocity(wec, x_wec, x_opt, waves, nsubsteps)
        update_force_td(vel_td)
    if integral:
        pos_td = pto.position(wec, x_wec, x_opt, waves, nsubsteps)
        update_force_td(pos_td)
    if derivative:
        acc_td = pto.acceleration(wec, x_wec, x_opt, waves, nsubsteps)
        update_force_td(acc_td)
    return force_td


def controller_pi(
    pto: TPTO,
    wec: TWEC,
    x_wec: ndarray,
    x_opt: ndarray,
    waves: Optional[Dataset] = None,
    nsubsteps: Optional[int] = 1,
) -> ndarray:
    """Proportional-integral (PI) controller that returns a time
    history of PTO forces.

    Parameters
    ----------
    pto
        :py:class:`wecopttool.pto.PTO` object.
    wec
        :py:class:`wecopttool.WEC` object.
    x_wec
        WEC dynamic state.
    x_opt
        Optimization (control) state.
    waves
        :py:class:`xarray.Dataset` with the structure and elements shown
        by :py:mod:`wecopttool.waves`.
    nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.
    """
    force_td = controller_pid(pto, wec, x_wec, x_opt, waves, nsubsteps,
                              True, True, False)
    return force_td


def controller_p(
    pto: TPTO,
    wec: TWEC,
    x_wec: ndarray,
    x_opt: ndarray,
    waves: Optional[Dataset] = None,
    nsubsteps: Optional[int] = 1,
) -> ndarray:
    """Proportional (P) controller that returns a time history of
    PTO forces.

    Parameters
    ----------
    pto
        :py:class:`wecopttool.pto.PTO` object.
    wec
        :py:class:`wecopttool.WEC` object.
    x_wec
        WEC dynamic state.
    x_opt
        Optimization (control) state.
    waves
        :py:class:`xarray.Dataset` with the structure and elements shown
        by :py:mod:`wecopttool.waves`.
    nsubsteps
        Number of steps between the default (implied) time steps.
        A value of :python:`1` corresponds to the default step length.
    """
    force_td = controller_pid(pto, wec, x_wec, x_opt, waves, nsubsteps,
                               True, False, False)
    return force_td



""" Unit tests for functions in the :python:`pto.py` module.
"""

import pytest
import numpy as np

import wecopttool as wot


# setup a simple PTO based on the WaveBot device
# TODO: replace with multi-DOF PTO
@pytest.fixture(scope="module")
def ndof():
    """Number of PTO degrees of freedom."""
    return 1


@pytest.fixture(scope="module")
def nfreq():
    """Number of frequencies (harmonics)."""
    return 2


@pytest.fixture(scope="module")
def f1():
    """Fundamental frequency [Hz]."""
    return 0.1


@pytest.fixture(scope="module")
def freq(f1, nfreq):
    """Frequency array [Hz]."""
    return np.arange(nfreq+1)*f1


@pytest.fixture(scope="module")
def omega(freq):
    """Radial frequency array [rad/s]."""
    return freq * 2*np.pi


@pytest.fixture(scope="module")
def pto_parameters():
    """Dictionary of PTO parameters."""
    params = {
        'gear_ratio': 12.0,  # [rad/m]
        'torque_constant': 6.7,  # [N*m/A]
        'winding_resistance': 0.5,  # [ohm]
        'winding_inductance': 0.0,  # [H]
        'drivetrain_inertia': 2.0,  # [kg*m^2]
        'drivetrain_friction': 1.0,  # [N*m*s/rad]
        'drivetrain_stiffness': 0.0,  # [N*m/rad]
    }
    return params


@pytest.fixture(scope="module")
def drivetrain_impedance(omega, pto_parameters):
    """Drive-train impedance."""
    imp = (
        1j*omega[1:]*pto_parameters['drivetrain_inertia'] +
        pto_parameters['drivetrain_friction'] +
        1/(1j*omega[1:])*pto_parameters['drivetrain_stiffness']
    )
    return imp


@pytest.fixture(scope="module")
def winding_impedance(omega, pto_parameters):
    """Generator winding impedance."""
    imp = (
        pto_parameters['winding_resistance'] +
        1j*omega[1:]*pto_parameters['winding_inductance']
    )
    return imp


@pytest.fixture(scope="module")
def pto_impedance(
        nfreq, pto_parameters, drivetrain_impedance, winding_impedance
    ):
    """PTO impedance matrix."""
    tc = pto_parameters['torque_constant']
    gr = pto_parameters['gear_ratio']
    # diagonals
    pto_impedance_11 = -1* gr**2 * drivetrain_impedance
    pto_impedance_22 = winding_impedance
    # off-diagonals
    off_diag = np.sqrt(3.0/2.0) * tc * gr
    pto_impedance_12 = -1*(off_diag+0j) * np.ones([nfreq])
    pto_impedance_21 = -1*(off_diag+0j) * np.ones([nfreq])
    # matrix
    imp = np.array([[pto_impedance_11, pto_impedance_12],
                    [pto_impedance_21, pto_impedance_22]])
    return imp


@pytest.fixture(scope="module")
def abcd(pto_impedance, ndof):
    return wot.pto._make_abcd(pto_impedance, ndof)


class TestSupportFunctions:
    """Test the hidden support functions in the :python:`pto` module."""

    def test_make_abcd(
            self, nfreq, pto_parameters, drivetrain_impedance,
            winding_impedance, abcd,
        ):
        """Test the function :python:`pto._make_abcd`"""
        tc = pto_parameters['torque_constant']
        gr = pto_parameters['gear_ratio']
        expected = np.zeros([2, 2, nfreq], complex)
        expected[0, 0, :] = -np.sqrt(2/3)*gr/tc*drivetrain_impedance
        expected[1, 1, :] = -np.sqrt(2/3)/gr/tc*winding_impedance
        expected[0, 1, :] = -np.sqrt(2/3)/gr/tc
        expected[1, 0, :] = (
            -np.sqrt(3/2)*gr*tc -
            np.sqrt(2/3)*gr/tc*winding_impedance*drivetrain_impedance
        )
        assert np.allclose(abcd, expected)

    def test_make_mimo_transfer_mat(self, abcd, ndof, nfreq):
        """Test the function :python:`pto._mimo_transfer_mat`"""
        mimo = wot.pto._make_mimo_transfer_mat(abcd, ndof)
        n = 2*nfreq
        expected = np.zeros([2*ndof*n, 2*ndof*n])
        # 0,0
        imp = abcd[0, 0, :]
        r0 = np.real(imp[0])
        i0 = np.imag(imp[0])
        r1 = np.real(imp[1])
        expected[:n, :n] = np.array([
            [r0, 0,    0, 0],
            [0, r0, -i0, 0],
            [0, i0,  r0, 0],
            [0, 0,    0, r1],
        ])
        # 0,1
        imp = abcd[0, 1, :]
        r0 = np.real(imp[0])
        i0 = np.imag(imp[0])
        r1 = np.real(imp[1])
        expected[:n, n:] = np.array([
            [r0, 0,    0, 0],
            [0, r0, -i0, 0],
            [0, i0,  r0, 0],
            [0, 0,    0, r1],
        ])
        # 1,0
        imp = abcd[1, 0, :]
        r0 = np.real(imp[0])
        i0 = np.imag(imp[0])
        r1 = np.real(imp[1])
        expected[n:, :n] = np.array([
            [r0, 0,    0, 0],
            [0, r0, -i0, 0],
            [0, i0,  r0, 0],
            [0, 0,    0, r1],
        ])
        # 1,1
        imp = abcd[1, 1, :]
        r0 = np.real(imp[0])
        i0 = np.imag(imp[0])
        r1 = np.real(imp[1])
        expected[n:, n:] = np.array([
            [r0, 0,    0, 0],
            [0, r0, -i0, 0],
            [0, i0,  r0, 0],
            [0, 0,    0, r1],
        ])
        # test
        assert np.allclose(mimo, expected)


class TestControllers:
    """Test the different sample controllers implemented in the
    :python:`pto` module."""

    @pytest.fixture(scope="class")
    def wec(self, f1, nfreq):
        """Empty WEC object."""
        return wot.WEC(f1, nfreq, {}, ndof=1, inertia_in_forces=True)

    @pytest.fixture(scope="class")
    def kinematics(self,):
        """PTO kinematics matrix."""
        return np.eye(1)

    @pytest.fixture(scope="class")
    def pid_p(self,):
        """PID controller proportional gains."""
        return 2.1

    @pytest.fixture(scope="class")
    def pid_i(self,):
        """PID controller integral gains."""
        return 3.4

    @pytest.fixture(scope="class")
    def pid_d(self,):
        """PID controller derivative gains."""
        return 2.1

    def test_controller_unstructured(self, wec, ndof, kinematics, omega):
        """Test the pseudo-spectral controller."""
        controller = wot.pto.controller_unstructured
        pto = wot.pto.PTO(ndof, kinematics, controller)
        amp = 1.2
        w = omega[-2]
        force = amp * np.cos(w * wec.time)
        force = force.reshape(-1, 1)
        x_opt = [0, amp, 0, 0]
        calculated = pto.force(wec, None, x_opt, None)
        assert np.allclose(force, calculated)

    def test_controller_p(self, wec, ndof, kinematics, omega, pid_p):
        """Test the proportional (P) controller."""
        controller = wot.pto.controller_p
        pto = wot.pto.PTO(ndof, kinematics, controller)
        amp = 2.3
        w = omega[-2]
        # pos = amp * np.cos(w * wec.time)
        vel = -1 * amp * w * np.sin(w * wec.time)
        force = vel*pid_p
        force = force.reshape(-1, 1)
        x_wec = [0, amp, 0, 0]
        x_opt = [pid_p,]
        calculated = pto.force(wec, x_wec, x_opt, None)
        assert np.allclose(force, calculated)

    def test_controller_pi(self, wec, ndof, kinematics, omega, pid_p, pid_i):
        """Test the PI controller."""
        controller = wot.pto.controller_pi
        pto = wot.pto.PTO(ndof, kinematics, controller)
        amp = 2.3
        w = omega[-2]
        pos = amp * np.cos(w * wec.time)
        vel = -1 * amp * w * np.sin(w * wec.time)
        force = vel*pid_p + pos*pid_i
        force = force.reshape(-1, 1)
        x_wec = [0, amp, 0, 0]
        x_opt = [pid_p, pid_i]
        calculated = pto.force(wec, x_wec, x_opt, None)
        assert np.allclose(force, calculated)

    def test_controller_pid(
            self, wec, ndof, kinematics, omega, pid_p, pid_i, pid_d,
        ):
        """Test the PID controller."""
        controller = wot.pto.controller_pid
        pto = wot.pto.PTO(ndof, kinematics, controller)
        amp = 2.3
        w = omega[-2]
        pos = amp * np.cos(w * wec.time)
        vel = -1 * amp * w * np.sin(w * wec.time)
        acc = -1 * amp * w**2 * np.cos(w * wec.time)
        force = vel*pid_p + pos*pid_i + acc*pid_d
        force = force.reshape(-1, 1)
        x_wec = [0, amp, 0, 0]
        x_opt = [pid_p, pid_i, pid_d]
        calculated = pto.force(wec, x_wec, x_opt, None)
        assert np.allclose(force, calculated)


"""Core functionality for solving the pseudo-spectral problem for WEC.

Contains:

* The *WEC* class
* Functions for basic functionality

.. note:: All contents of this module are imported into *WecOpTool* and
          can be called directly as :python:`wecopttool.<function>`
          instead of :python:`wecopttool.core.<function>`.
"""


from __future__ import annotations


__all__ = [
    "WEC",
    "ncomponents",
    "frequency",
    "time",
    "time_mat",
    "derivative_mat",
    "derivative2_mat",
    "mimo_transfer_mat",
    "vec_to_dofmat",
    "dofmat_to_vec",
    "real_to_complex",
    "complex_to_real",
    "fd_to_td",
    "td_to_fd",
    "read_netcdf",
    "write_netcdf",
    "check_linear_damping",
    "check_impedance",
    "force_from_rao_transfer_function",
    "force_from_impedance",
    "force_from_waves",
    "inertia",
    "standard_forces",
    "run_bem",
    "change_bem_convention",
    "linear_hydrodynamics",
    "wave_excitation",
    "hydrodynamic_impedance",
    "atleast_2d",
    "degrees_to_radians",
    "subset_close",
    "scale_dofs",
    "decompose_state",
    "frequency_parameters",
    "time_results",
]


import logging
from typing import Iterable, Callable, Any, Optional, Mapping, TypeVar, Union
from pathlib import Path
import warnings
from datetime import datetime

from numpy.typing import ArrayLike
import autograd.numpy as np
from autograd.numpy import ndarray
from autograd.builtins import isinstance, tuple, list, dict
from autograd import grad, jacobian
import xarray as xr
from xarray import DataArray, Dataset
import capytaine as cpy
from scipy.optimize import minimize, OptimizeResult, Bounds
from scipy.linalg import block_diag, dft


# logger
_log = logging.getLogger(__name__)

# autograd warnings
filter_msg = "Casting complex values to real discards the imaginary part"
warnings.filterwarnings("ignore", message=filter_msg)

# default values
_default_parameters = {'rho': 1025.0, 'g': 9.81, 'depth': np.infty}
_default_min_damping = 1e-6

# type aliases
TWEC = TypeVar("TWEC", bound="WEC")
TStateFunction = Callable[
    [TWEC, ndarray, ndarray, Dataset], ndarray]
TForceDict = dict[str, TStateFunction]
TIForceDict = Mapping[str, TStateFunction]
FloatOrArray = Union[float, ArrayLike]


class WEC:
    """A wave energy converter (WEC) object for performing simulations
    using the pseudo-spectral solution method.

    To create the WEC use one of the initialization methods:

    * :py:meth:`wecopttool.WEC.__init__`
    * :py:meth:`wecopttool.WEC.from_bem`
    * :py:meth:`wecopttool.WEC.from_floating_body`
    * :py:meth:`wecopttool.WEC.from_impedance`.

    .. note:: Direct initialization of a :py:class:`wecopttool.WEC`
        object as :python:`WEC(f1, nfreq, forces, ...)` using
        :py:meth:`wecopttool.WEC.__init__` is discouraged. Instead
        use one of the other initialization methods listed in the
        *See Also* section.

    To solve the pseudo-spectral problem use
    :py:meth:`wecopttool.WEC.solve`.
    """
    def __init__(
        self,
        f1: float,
        nfreq: int,
        forces: TIForceDict,
        constraints: Optional[Iterable[Mapping]] = None,
        inertia_matrix: Optional[ndarray] = None,
        ndof: Optional[int] = None,
        inertia_in_forces: Optional[bool] = False,
        dof_names: Optional[Iterable[str]] = None,
        ) -> None:
        """Create a WEC object directly from its inertia matrix and
        list of forces.

        The :py:class:`wecopttool.WEC` class describes a WEC's
        equation of motion as :math:`ma=Σf` where the
        :python:`inertia_matrix` matrix specifies the inertia :math:`m`,
        and the :python:`forces` dictionary specifies the different
        forces to be summed. The forces can be linear or nonlinear.
        If :python:`inertia_in_forces is True` the equation of motion is
        :math:`Σf=0`, which is included to allow for initialization
        using an intrinsic impedance through the
        :python:`WEC.from_impedance` initialization function.

        .. note:: Direct initialization of a
            :py:class:`wecopttool.WEC` object as
            :python:`WEC(f1, nfreq, forces, ...)` is discouraged.
            Instead use one of the other initialization methods listed
            in the *See Also* section.

        Parameters
        ----------
        f1
            Fundamental frequency :python:`f1` [:math:`Hz`].
        nfreq
            Number of frequencies (not including zero frequency),
            i.e., :python:`freqs = [0, f1, 2*f1, ..., nfreq*f1]`.
        forces
            Dictionary with entries :python:`{'force_name': fun}`,
            where :python:`fun` has a  signature
            :python:`def fun(wec, x_wec, x_opt, waves):`, and returns
            forces in the time-domain of size
            :python:`(2*nfreq+1, ndof)`.
        constraints
            List of constraints, see documentation for
            :py:func:`scipy.optimize.minimize` for description and
            options of constraints dictionaries.
            If :python:`None`: empty list :python:`[]`.
        inertia_matrix
           Inertia matrix of size :python:`(ndof, ndof)`.
           Not used if :python:`inertia_in_forces` is :python:`True`.
        ndof
            Number of degrees of freedom.
            Must be specified if :python:`inertia_in_forces is True`,
            else not used.
        inertia_in_forces
            Set to True if inertial "forces" are included in the
            :python:`forces` argument.
            This scenario is rare.
            If using an intrinsic impedance, consider initializing with
            :py:meth:`wecoptool.core.WEC.from_impedance` instead.
        dof_names
            Names of the different degrees of freedom (e.g.
            :python:`'Heave'`).
            If :python:`None` the names
            :python:`['DOF_0', ..., 'DOF_N']` are used.

        Raises
        ------
        ValueError
            If :python:`inertia_in_forces is True` but :python:`ndof` is
            not specified.
        ValueError
            If :python:`inertia_in_forces is False` but
            :python:`inertia_matrix` is not specified.
        ValueError
            If :python:`inertia_matrix` does not have the correct size
            (:python:`(ndof, ndof)`).
        ValueError
            If :python:`dof_names` does not have the correct size
            (:python:`ndof`).

        See Also
        --------
        from_bem:
            Initialize a :py:class:`wecopttool.WEC` object from BEM
            results.
        from_floating_body:
            Initialize a :py:class:`wecopttool.WEC` object from a
            :py:class:`capytaine.bodies.bodies.FloatingBody` object.
        from_impedance:
            Initialize a :py:class:`wecopttool.WEC` object from an
            intrinsic impedance array and excitation coefficients.
        """
        self._freq = frequency(f1, nfreq)
        self._time = time(f1, nfreq)
        self._time_mat = time_mat(f1, nfreq)
        self._derivative_mat = derivative_mat(f1, nfreq)
        self._derivative2_mat = derivative2_mat(f1, nfreq)
        self._forces = forces
        constraints = list(constraints) if (constraints is not None) else []
        self._constraints = constraints

        # inertia options
        self._inertia_in_forces = inertia_in_forces

        def _missing(var_name, condition):
            msg = (f"'{var_name}' must be provided if 'inertia_in_forces' is" +
                    f"'{condition}'.")
            return msg

        def _ignored(var_name, condition):
            msg = (f"'{var_name}' is not used when 'inertia_in_forces' is " +
                    f"'{condition}' and should not be provided")
            return msg

        if inertia_in_forces:
            condition = "True"
            if inertia_matrix is not None:
                _log.warning(_ignored("inertia_matrix", condition))
                inertia_matrix = None
            if ndof is None:
                raise ValueError(_missing("ndof", condition))
        elif not inertia_in_forces:
            condition = "False"
            if inertia_matrix is None:
                raise ValueError(_missing("inertia_matrix", condition))
            inertia_matrix = np.atleast_2d(np.squeeze(inertia_matrix))
            if ndof is not None:
                _log.warning(_ignored("ndof", condition))
                if ndof != inertia_matrix.shape[0]:
                    _log.warning(
                        "Provided value of 'ndof' does not match size of " +
                        "'inertia_matrix'. Setting " +
                        f"'ndof={inertia_matrix.shape[0]}'.")
            ndof = inertia_matrix.shape[0]

            if inertia_matrix.shape != (ndof, ndof):
                raise ValueError(
                    "'inertia_matrix' must be a square matrix of size equal " +
                    "to the number of degrees of freedom.")
        self._inertia_matrix = inertia_matrix
        self._ndof = ndof
        if inertia_in_forces:
            _inertia = None
        else:
            _inertia = inertia(f1, nfreq, inertia_matrix)
        self._inertia = _inertia

        # names
        if dof_names is None:
            dof_names = [f'DOF_{i}' for i in range(ndof)]
        elif len(dof_names) != ndof:
            raise ValueError("'dof_names' must have length 'ndof'.")
        self._dof_names = list(dof_names)

    def __str__(self) -> str:
        str = (f'{self.__class__.__name__}: ' +
               f'DOFs ({self.ndof})={self.dof_names}, ' +
               f'f=[0, {self.f1}, ..., {self.nfreq}({self.f1})] Hz.')
        return str

    def __repr__(self) -> str:
        type_ = type(self)
        module = type_.__module__
        qualname = type_.__qualname__
        repr_org = f"<{module}.{qualname} object at {hex(id(self))}>"
        return repr_org + " :: " + self.__str__()

    # other initialization methods
    @staticmethod
    def from_bem(
        bem_data: Union[Dataset, Union[str, Path]],
        inertia_matrix: Optional[ndarray] = None,
        hydrostatic_stiffness: Optional[ndarray] = None,
        friction: Optional[ndarray] = None,
        f_add: Optional[TIForceDict] = None,
        constraints: Optional[Iterable[Mapping]] = None,
        min_damping: Optional[float] = _default_min_damping,
        uniform_shift: Optional[bool] = True,
        dof_names: Optional[Iterable[str]] = None,
        ) -> TWEC:
        """Create a WEC object from linear hydrodynamic coefficients
        obtained using the boundary element method (BEM) code Capytaine.

        The :python:`bem_data` can be a dataset or the name of a
        *NetCDF* file containing the dataset.

        The returned :py:class:`wecopttool.WEC` object contains the
        inertia and the default linear forces: radiation, diffraction,
        and Froude-Krylov. Additional forces can be specified through
        :python:`f_add`.

        Note that because Capytaine uses a different sign convention,
        the direct results from capytaine must be modified using
        :py:func:`wecopttool.change_bem_convention` before calling
        this initialization function.
        Instead, the recommended approach is to use
        :py:func:`wecopttool.run_bem`,
        rather than running Capytaine directly, which outputs the
        results in the correct convention. The results can be saved
        using :py:func:`wecopttool.write_netcdf`.

        In addition to the Capytaine results, if the dataset contains
        the :python:`inertia_matrix`, :python:`hydrostatic_stiffness`,
        or :python:`friction` these do not need to be provided
        separately.

        Parameters
        ----------
        bem_data
            Linear hydrodynamic coefficients obtained using the boundary
            element method (BEM) code Capytaine, with sign convention
            corrected.
        inertia_matrix
           Inertia matrix of size :python:`(ndof, ndof)`.
           :python:`None` if included in :python:`bem_data`.
        hydrostatic_stiffness
            Linear hydrostatic restoring coefficient of size
            :python:`(nodf, ndof)`.
            :python:`None` if included in :python:`bem_data`.
        friction
            Linear friction, in addition to radiation damping, of size
            :python:`(nodf, ndof)`.
            :python:`None` if included in :python:`bem_data` or to set
            to zero.
        f_add
            Dictionary with entries :python:`{'force_name': fun}`, where
            :python:`fun` has a  signature
            :python:`def fun(wec, x_wec, x_opt, waves):`, and returns
            forces in the time-domain of size
            :python:`(2*nfreq+1, ndof)`.
        constraints
            List of constraints, see documentation for
            :py:func:`scipy.optimize.minimize` for description and
            options of constraints dictionaries.
            If :python:`None`: empty list :python:`[]`.
        min_damping
            Minimum damping level to ensure a stable system.
            See :py:func:`wecopttool.check_linear_damping` for more details.
        uniform_shift
            Boolean determining whether damping corrections shifts the damping
            values uniformly for all frequencies or only for frequencies below
            :python:`min_damping`.
            See :py:func:`wecopttool.check_linear_damping` for more details.
        dof_names
            Names of the different degrees of freedom (e.g.
            :python:`'Heave'`).
            If :python:`None` the names
            :python:`['DOF_0', ..., 'DOF_N']` are used.

        Raises
        ------
        ValueError
            If either :python:`inertia_matrix` or
            :python:`hydrostatic_stiffness` are :python:`None` and is
            not included in :python:`bem_data`.
            See :py:func:`wecopttool.linear_hydrodynamics`.
        ValueError
            If any of :python:`inertia_matrix`,
            :python:`hydrostatic_stiffness`, or :python:`stiffness` are
            both provided and included in :python:`bem_data` but have
            different values.
            See :py:func:`wecopttool.linear_hydrodynamics`.

        See Also
        --------
        run_bem, linear_hydrodynamics, change_bem_convention,
        write_netcdf, check_linear_damping
        """
        if isinstance(bem_data, (str, Path)):
            bem_data = read_netcdf(bem_data)
        # add inertia_matrix, hydrostatic stiffness, and friction
        hydro_data = linear_hydrodynamics(
            bem_data, inertia_matrix, hydrostatic_stiffness, friction)
        if inertia_matrix is None:
            inertia_matrix = hydro_data['inertia_matrix'].values

        # frequency array
        f1, nfreq = frequency_parameters(
            hydro_data.omega.values/(2*np.pi), False)

        # check real part of damping diagonal > 0
        if min_damping is not None:
            hydro_data = check_linear_damping(
                hydro_data, min_damping, uniform_shift)

        # forces in the dynamics equations
        linear_force_functions = standard_forces(hydro_data)
        f_add = f_add if (f_add is not None) else {}
        forces = linear_force_functions | f_add
        # constraints
        constraints = constraints if (constraints is not None) else []
        return WEC(f1, nfreq, forces, constraints, inertia_matrix, dof_names=dof_names)

    @staticmethod
    def from_floating_body(
        fb: cpy.FloatingBody,
        f1: float,
        nfreq: int,
        inertia_matrix: ndarray,
        hydrostatic_stiffness: ndarray,
        friction: Optional[ndarray] = None,
        f_add: Optional[TIForceDict] = None,
        constraints: Optional[Iterable[Mapping]] = None,
        min_damping: Optional[float] = _default_min_damping,
        wave_directions: Optional[ArrayLike] = np.array([0.0,]),
        rho: Optional[float] = _default_parameters['rho'],
        g: Optional[float] = _default_parameters['g'],
        depth: Optional[float] = _default_parameters['depth'],
    ) -> TWEC:
        """Create a WEC object from a Capytaine :python:`FloatingBody`
        (:py:class:capytaine.bodies.bodies.FloatingBody).

        A :py:class:capytaine.bodies.bodies.FloatingBody object contains
        information on the mesh and degrees of freedom.

        This initialization method calls :py:func:`wecopttool.run_bem`
        followed by :py:meth:`wecopttool.WEC.from_bem`.

        This will run Capytaine to obtain the linear hydrodynamic
        coefficients, which can take from a few minutes to several
        hours.
        Instead, if the hydrodynamic coefficients can be reused, it is
        recommended to run Capytaine first and save the results using
        :py:func:`wecopttool.run_bem` and
        :py:func:`wecopttool.write_netcdf`,
        and then initialize the :py:class:`wecopttool.WEC` object
        using :py:meth:`wecopttool.WEC.from_bem`.
        This initialization method should be
        reserved for the cases where the hydrodynamic coefficients
        constantly change and are not reused, as for example for
        geometry optimization.

        Parameters
        ----------
        fb
            Capytaine FloatingBody.
        f1
            Fundamental frequency :python:`f1` [:math:`Hz`].
        nfreq
            Number of frequencies (not including zero frequency),
            i.e., :python:`freqs = [0, f1, 2*f1, ..., nfreq*f1]`.
        inertia_matrix
           Inertia matrix of size :python:`(ndof, ndof)`.
        hydrostatic_stiffness
            Linear hydrostatic restoring coefficient of size
            :python:`(ndof, ndof)`.
        friction
            Linear friction, in addition to radiation damping, of size
            :python:`(ndof, ndof)`.
            :python:`None` to set to zero.
        f_add
            Dictionary with entries :python:`{'force_name': fun}`, where
            :python:`fun` has a  signature
            :python:`def fun(wec, x_wec, x_opt, waves):`, and returns
            forces in the time-domain of size
            :python:`(2*nfreq, ndof)`.
        constraints
            List of constraints, see documentation for
            :py:func:`scipy.optimize.minimize` for description and
            options of constraints dictionaries.
            If :python:`None`: empty list :python:`[]`.
        min_damping
            Minimum damping level to ensure a stable system.
            See :py:func:`wecopttool.check_linear_damping` for
            more details.
        wave_directions
            List of wave directions [degrees] to evaluate BEM at.
        rho
            Water density in :math:`kg/m^3`.
        g
            Gravitational acceleration in :math:`m/s^2`.
        depth
            Water depth in :math:`m`.

        Returns
        -------
        WEC
            An instance of the :py:class:`wecopttool.WEC` class.

        See Also
        --------
        run_bem, write_netcdf, WEC.from_bem
        """

        # RUN BEM
        _log.info(f"Running Capytaine (BEM): {nfreq+1} frequencies x " +
                 f"{len(wave_directions)} wave directions.")
        freq = frequency(f1, nfreq)[1:]
        bem_data = run_bem(
            fb, freq, wave_directions, rho=rho, g=g, depth=depth)
        wec = WEC.from_bem(
            bem_data, inertia_matrix, hydrostatic_stiffness, friction, f_add,
            constraints, min_damping=min_damping)
        return wec

    @staticmethod
    def from_impedance(
        freqs: ArrayLike,
        impedance: ArrayLike,
        exc_coeff: ArrayLike,
        hydrostatic_stiffness: ndarray,
        f_add: Optional[TIForceDict] = None,
        constraints: Optional[Iterable[Mapping]] = None,
        min_damping: Optional[float] = _default_min_damping,
    ) -> TWEC:
        """Create a WEC object from the intrinsic impedance and
        excitation coefficients.

        The intrinsic (mechanical) impedance :math:`Z(ω)` linearly
        relates excitation forces :math:`F(ω)` to WEC velocity
        :math:`U(ω)` as :math:`ZU=F`.
        Using linear hydrodynamic coefficients, e.g. from a BEM code
        like Capytaine, the impedance is given as
        :math:`Z(ω) = (m+A(ω))*iω + B(ω) + B_f + K/(iω)`.
        The impedance can also be obtained experimentally.
        Note that the impedance is not defined at :math:`ω=0`.


        Parameters
        ----------
        freqs
            Frequency vector [:math:`Hz`] not including the zero frequency,
            :python:`freqs = [f1, 2*f1, ..., nfreq*f1]`.
        impedance
            Complex impedance of size :python:`(ndof, ndof, nfreq)`.
        exc_coeff
            Complex excitation transfer function of size
            :python:`(ndof, nfreq)`.
        hydrostatic_stiffness
            Linear hydrostatic restoring coefficient of size
            :python:`(ndof, ndof)`.
        f_add
            Dictionary with entries :python:`{'force_name': fun}`, where
            :python:`fun` has a  signature
            :python:`def fun(wec, x_wec, x_opt, waves):`, and returns
            forces in the time-domain of size
            :python:`(2*nfreq, ndof)`.
        constraints
            List of constraints, see documentation for
            :py:func:`scipy.optimize.minimize` for description and
            options of constraints dictionaries.
            If :python:`None`: empty list :python:`[]`.
        min_damping
            Minimum damping level to ensure a stable system.
            See :py:func:`wecopttool.check_impedance` for
            more details.

        Raises
        ------
        ValueError
            If :python:`impedance` does not have the correct size:
            :python:`(ndof, ndof, nfreq)`.
        """
        f1, nfreq = frequency_parameters(freqs, False)

        # impedance matrix shape
        shape = impedance.shape
        ndim = impedance.ndim
        if (ndim!=3) or (shape[0]!=shape[1]) or (shape[2]!=nfreq):
            raise ValueError(
                "'impedance' must have shape '(ndof, ndof, nfreq)'.")

        impedance = check_impedance(impedance, min_damping)

        # impedance force
        omega = freqs * 2*np.pi
        transfer_func = impedance * (1j*omega)
        transfer_func0 = np.expand_dims(hydrostatic_stiffness, 2)
        transfer_func = np.concatenate([transfer_func0, transfer_func], 2)
        transfer_func = -1 * transfer_func  # RHS of equation: ma = Σf
        force_impedance = force_from_rao_transfer_function(transfer_func)

        # excitation force
        force_excitation = force_from_waves(exc_coeff)

        # all forces
        f_add = {} if (f_add is None) else f_add
        forces =  {
            'intrinsic_impedance': force_impedance,
            'excitation': force_excitation
        }
        forces = forces | f_add

        # wec
        wec = WEC(f1, nfreq, forces, constraints,
                  inertia_in_forces=True, ndof=shape[0])
        return wec

    def _resid_fun(self, x_wec, x_opt, waves):
        if not self.inertia_in_forces:
            ri = self.inertia(self, x_wec, x_opt, waves)
        else:
            ri = np.zeros([self.ncomponents, self.ndof])
        # forces, -Σf
        for f in self.forces.values():
            ri = ri - f(self, x_wec, x_opt, waves)
        return self.dofmat_to_vec(ri)

    # solve
    def solve(self,
        waves: Dataset,
        obj_fun: TStateFunction,
        nstate_opt: int,
        x_wec_0: Optional[ndarray] = None,
        x_opt_0: Optional[ndarray] = None,
        scale_x_wec: Optional[list] = None,
        scale_x_opt: Optional[FloatOrArray] = 1.0,
        scale_obj: Optional[float] = 1.0,
        optim_options: Optional[Mapping[str, Any]] = {},
        use_grad: Optional[bool] = True,
        maximize: Optional[bool] = False,
        bounds_wec: Optional[Bounds] = None,
        bounds_opt: Optional[Bounds] = None,
        callback: Optional[TStateFunction] = None,
        ) -> tuple[Dataset, Dataset, OptimizeResult]:
        """Simulate WEC dynamics using a pseudo-spectral solution
        method and returns the raw results dictionary produced by 
        :py:func:`scipy.optimize.minimize`.

        Parameters
        ----------
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        obj_fun
            Objective function to minimize for pseudo-spectral solution,
            must have signature :python:`fun(wec, x_wec, x_opt, waves)`
            and return a scalar.
        nstate_opt
            Length of the optimization (controls) state vector.
        x_wec_0
            Initial guess for the WEC dynamics state.
            If :python:`None` it is randomly initiated.
        x_opt_0
            Initial guess for the optimization (control) state.
            If :python:`None` it is randomly initiated.
        scale_x_wec
            Factor(s) to scale each DOF in :python:`x_wec` by, to
            improve convergence.
            A single float or an array of size :python:`ndof`.
        scale_x_opt
            Factor(s) to scale :python:`x_opt` by, to improve
            convergence.
            A single float or an array of size :python:`nstate_opt`.
        scale_obj
            Factor to scale :python:`obj_fun` by, to improve
            convergence.
        optim_options
            Optimization options passed to the optimizer.
            See :py:func:`scipy.optimize.minimize`.
        use_grad
             If :python:`True`, optimization will utilize
             `autograd <https://github.com/HIPS/autograd>`_
             for gradients.
        maximize
            Whether to maximize the objective function.
            The default is to minimize the objective function.
        bounds_wec
            Bounds on the WEC components of the decision variable.
            See :py:func:`scipy.optimize.minimize`.
        bounds_opt
            Bounds on the optimization (control) components of the
            decision variable.
            See :py:func:`scipy.optimize.minimize`.
        callback
            Function called after each iteration, must have signature
            :python:`fun(wec, x_wec, x_opt, waves)`. The default
            provides status reports at each iteration via logging at the
            INFO level.

        Raises
        ------
        ValueError
            If :python:`scale_x_opt` is a scalar and
            :python:`nstate_opt` is not provided.
        Exception
            If the optimizer fails for any reason other than maximum
            number of states, i.e. for exit modes other than 0 or 9.
            See :py:mod:`scipy.optimize` for exit mode details.
            
        Examples
        --------
        The :py:meth:`wecopttool.WEC.solve` method only returns the
        raw results dictionary produced by :py:func:`scipy.optimize.minimize`.

        >>> res_opt = wec.solve(waves=wave,
                                obj_fun=pto.average_power,
                                nstate_opt=2*nfreq+1)

        To get the post-processed results for the
        :py:class:`wecopttool.pto.PTO`, you may call

        >>> res_wec_fd, res_wec_td = wec.post_process(wec,res_opt)
        >>> res_pto_fd, res_pto_td = pto.post_process(wec,res_opt)

        See Also
        --------
        wecopttool.waves,
        """

        _log.info("Solving pseudo-spectral control problem.")

        # x_wec scaling vector
        if scale_x_wec == None:
            scale_x_wec = [1.0] * self.ndof
        elif isinstance(scale_x_wec, float) or isinstance(scale_x_wec, int):
            scale_x_wec = [scale_x_wec] * self.ndof
        scale_x_wec = scale_dofs(scale_x_wec, self.ncomponents)

        # x_opt scaling vector
        if isinstance(scale_x_opt, float) or isinstance(scale_x_opt, int):
            if nstate_opt is None:
                raise ValueError("If 'scale_x_opt' is a scalar, " +
                                    "'nstate_opt' must be provided")
            scale_x_opt = scale_dofs([scale_x_opt], nstate_opt)

        # composite scaling vector
        scale = np.concatenate([scale_x_wec, scale_x_opt])

        # decision variable initial guess
        if x_wec_0 is None:
            x_wec_0 = np.random.randn(self.nstate_wec)
        if x_opt_0 is None:
            x_opt_0 = np.random.randn(nstate_opt)
        x0 = np.concatenate([x_wec_0, x_opt_0])*scale

        # objective function
        sign = -1.0 if maximize else 1.0

        def obj_fun_scaled(x):
            x_wec, x_opt = self.decompose_state(x/scale)
            return obj_fun(self, x_wec, x_opt, waves)*scale_obj*sign

        # constraints
        constraints = self.constraints.copy()

        for i, icons in enumerate(self.constraints):
            icons_new = {"type": icons["type"]}

            def make_new_fun(icons):
                def new_fun(x):
                    x_wec, x_opt = self.decompose_state(x/scale)
                    return icons["fun"](self, x_wec, x_opt, waves)
                return new_fun

            icons_new["fun"] = make_new_fun(icons)
            if use_grad:
                icons_new['jac'] = jacobian(icons_new['fun'])
            constraints[i] = icons_new

        # system dynamics through equality constraint, ma - Σf = 0
        def scaled_resid_fun(x):
            x_s = x/scale
            x_wec, x_opt = self.decompose_state(x_s)
            return self._resid_fun(x_wec, x_opt, waves)

        eq_cons = {'type': 'eq', 'fun': scaled_resid_fun}
        if use_grad:
            eq_cons['jac'] = jacobian(scaled_resid_fun)
        constraints.append(eq_cons)

        # bounds
        if (bounds_wec is None) and (bounds_opt is None):
            bounds = None
        else:
            bounds_in = [bounds_wec, bounds_opt]
            for idx, bii in enumerate(bounds_in):
                if isinstance(bii, tuple):
                    bounds_in[idx] = Bounds(lb=[xibs[0] for xibs in bii],
                                            ub=[xibs[1] for xibs in bii])
            inf_wec = np.ones(self.nstate_wec)*np.inf
            inf_opt = np.ones(nstate_opt)*np.inf
            bounds_dflt = [Bounds(lb=-inf_wec, ub=inf_wec),
                            Bounds(lb=-inf_opt, ub=inf_opt)]
            bounds_list = []
            for bi, bd in zip(bounds_in, bounds_dflt):
                if bi is not None:
                    bo = bi
                else:
                    bo = bd
                bounds_list.append(bo)
            bounds = Bounds(lb=np.hstack([le.lb for le in bounds_list])*scale,
                            ub=np.hstack([le.ub for le in bounds_list])*scale)

        # callback
        if callback is None:
            def callback_scipy(x):
                x_wec, x_opt = self.decompose_state(x)
                _log.info("[max(x_wec), max(x_opt), obj_fun(x)]: "
                          + f"[{np.max(np.abs(x_wec)):.2e}, "
                          + f"{np.max(np.abs(x_opt)):.2e}, "
                          + f"{np.max(obj_fun_scaled(x)):.2e}]")
        else:
            def callback_scipy(x):
                x_wec, x_opt = self.decompose_state(x)
                return callback(self, x_wec, x_opt, waves)

        # optimization problem
        optim_options['disp'] = optim_options.get('disp', True)
        problem = {'fun': obj_fun_scaled,
                    'x0': x0,
                    'method': 'SLSQP',
                    'constraints': constraints,
                    'options': optim_options,
                    'bounds': bounds,
                    'callback': callback_scipy,
                    }
        if use_grad:
            problem['jac'] = grad(obj_fun_scaled)

        # minimize
        optim_res = minimize(**problem)

        msg = f'{optim_res.message}    (Exit mode {optim_res.status})'
        if optim_res.status == 0:
            _log.info(msg)
        elif optim_res.status == 9:
            _log.warning(msg)
        else:
            raise Exception(msg)

        # unscale
        optim_res.x = optim_res.x / scale
        optim_res.fun = optim_res.fun / scale_obj
        optim_res.jac = optim_res.jac / scale_obj * scale

        return optim_res

    def post_process(self,
        res: OptimizeResult,
        waves: Dataset,
        nsubsteps: Optional[int] = 1,
    ) -> tuple[Dataset, Dataset]:
        """Post-process the results from
        :py:meth:`wecopttool.WEC.solve`.

        Parameters
        ----------
        waves
            :py:class:`xarray.Dataset` with the structure and elements
            shown by :py:mod:`wecopttool.waves`.
        res
            Results produced by :py:func:`scipy.optimize.minimize`.
        nsubsteps
            Number of steps between the default (implied) time steps.
            A value of :python:`1` corresponds to the default step
            length.

        Returns
        -------
        results_fd
            Dynamic responses in the frequency-domain.
        results_td
            Dynamic responses in the time-domain.
            
        Examples
        --------
        The :py:meth:`wecopttool.WEC.solve` method only returns the
        raw results dictionary produced by :py:func:`scipy.optimize.minimize`.

        >>> res_opt = wec.solve(waves=wave,
                                obj_fun=pto.average_power,
                                nstate_opt=2*nfreq+1)

        To get the post-processed results for the
        :py:class:`wecopttool.pto.PTO`, you may call

        >>> res_wec_fd, res_wec_td = wec.post_process(wec,res_opt)
        >>> res_pto_fd, res_pto_td = pto.post_process(wec,res_opt)
        """
        create_time = f"{datetime.utcnow()}"

        pos_attr = {'long_name': 'Position', 'units': 'm or rad'}
        vel_attr = {'long_name': 'Velocity', 'units': 'm/s or rad/s'}
        acc_attr = {'long_name': 'Acceleration', 'units': 'm/s^2 or rad/s^2'}
        omega_attr = {'long_name': 'Radial frequency', 'units': 'rad/s'}
        freq_attr = {'long_name': 'Frequency', 'units': 'Hz'}
        period_attr = {'long_name': 'Period', 'units': 's'}
        time_attr = {'long_name': 'Time', 'units': 's'}
        dof_attr = {'long_name': 'Degree of freedom'}
        force_attr = {'long_name': 'Force or moment', 'units': 'N or Nm'}
        wave_elev_attr = {'long_name': 'Wave elevation', 'units': 'm'}
        x_wec, x_opt = self.decompose_state(res.x)
        omega_coord = ("omega", self.omega, omega_attr)
        freq_coord = ("omega", self.frequency, freq_attr)
        period_coord = ("omega", self.period, period_attr)
        dof_coord = ("influenced_dof", self.dof_names, dof_attr)

        # frequency domain
        force_da_list = []
        for name, force in self.forces.items():
            force_td_tmp = force(self, x_wec, x_opt, waves)
            force_fd = self.td_to_fd(force_td_tmp)
            force_da = DataArray(data=force_fd,
                                 dims=["omega", "influenced_dof"],
                                 coords={
                                     'omega': omega_coord,
                                     'freq': freq_coord,
                                     'period': period_coord,
                                     'influenced_dof': dof_coord},
                                 attrs=force_attr
                                 ).expand_dims({'type': [name]})
            force_da_list.append(force_da)

        fd_forces = xr.concat(force_da_list, dim='type')
        fd_forces.type.attrs['long_name'] = 'Type'
        fd_forces.name = 'force'
        fd_forces.attrs['long_name'] = 'Force'

        pos = self.vec_to_dofmat(x_wec)
        pos_fd = real_to_complex(pos)

        vel = self.derivative_mat @ pos
        vel_fd = real_to_complex(vel)

        acc = self.derivative2_mat @ pos
        acc_fd = real_to_complex(acc)

        fd_state = Dataset(
            data_vars={
                'pos': (['omega', 'influenced_dof'], pos_fd, pos_attr),
                'vel': (['omega', 'influenced_dof'], vel_fd, vel_attr),
                'acc': (['omega', 'influenced_dof'], acc_fd, acc_attr)},
            coords={
                'omega': omega_coord,
                'freq': freq_coord,
                'period': period_coord,
                'influenced_dof': dof_coord},
            attrs={"time_created_utc": create_time}
        )

        results_fd = xr.merge([fd_state, fd_forces, waves])
        results_fd = results_fd.transpose('omega', 'influenced_dof', 'type',
                                          'wave_direction')
        results_fd = results_fd.fillna(0)

        # time domain
        t_dat = self.time_nsubsteps(nsubsteps)
        time = DataArray(
            data=t_dat, name='time', dims='time', coords=[t_dat])
        results_td = results_fd.map(lambda x: time_results(x, time))

        results_td['pos'].attrs = pos_attr
        results_td['vel'].attrs = vel_attr
        results_td['acc'].attrs = acc_attr
        results_td['wave_elev'].attrs = wave_elev_attr
        results_td['force'].attrs = force_attr
        results_td['time'].attrs = time_attr
        results_td.attrs['time_created_utc'] = create_time

        return results_fd, results_td

    # properties
    @property
    def forces(self) -> TForceDict:
        """Dictionary of forces."""
        return self._forces

    @forces.setter
    def forces(self, val):
        self._forces = dict(val)

    @property
    def constraints(self) -> list[dict]:
        """List of constraints."""
        return self._constraints

    @constraints.setter
    def constraints(self, val):
        self._constraints = list(val)

    @property
    def inertia_in_forces(self) -> bool:
        """Whether inertial "forces" are included in the
        :python:`forces` dictionary.
        """
        return self._inertia_in_forces

    @property
    def inertia_matrix(self) -> ndarray:
        """Inertia (mass) matrix.
        :python:`None` if  :python:`inertia_in_forces is True`.
        """
        return self._inertia_matrix

    @property
    def inertia(self) -> TStateFunction:
        """Function representing the inertial term :math:`ma` in the
        WEC's dynamics equation.
        """
        return self._inertia

    @property
    def dof_names(self) -> list[str]:
        """Names of the different degrees of freedom."""
        return self._dof_names

    @property
    def ndof(self) -> int:
        """Number of degrees of freedom."""
        return self._ndof

    @property
    def frequency(self) -> ndarray:
        """Frequency vector [:math:`Hz`]."""
        return self._freq

    @property
    def f1(self) -> float:
        """Fundamental frequency :python:`f1` [:math:`Hz`]."""
        return self._freq[1]

    @property
    def nfreq(self) -> int:
        """Number of frequencies, not including the zero-frequency."""
        return len(self._freq)-1

    @property
    def omega(self) -> ndarray:
        """Radial frequency vector [rad/s]."""
        return self._freq * (2*np.pi)

    @property
    def period(self) -> ndarray:
        """Period vector [s]."""
        return np.concatenate([[np.Infinity], 1/self._freq[1:]])

    @property
    def w1(self) -> float:
        """Fundamental radial frequency [rad/s]."""
        return self.omega[1]

    @property
    def time(self) -> ndarray:
        """Time vector [s], size '(2*nfreq, ndof)', not containing the
        end time 'tf'."""
        return self._time

    @property
    def time_mat(self) -> ndarray:
        """Matrix to create time-series from Fourier coefficients.

        For some array of Fourier coefficients :python:`x`
        (excluding the sine component of the highest freequency), size
        :python:`(2*nfreq, ndof)`, the time series is obtained via
        :python:`time_mat @ x`, also size
        :python:`(2*nfreq, ndof)`.
        """
        return self._time_mat

    @property
    def derivative_mat(self) -> ndarray:
        """Matrix to create Fourier coefficients of the derivative of
        some quantity.

        For some array of Fourier coefficients :python:`x`
        (excluding the sine component of the highest freequency), size
        :python:`(2*nfreq, ndof)`, the Fourier coefficients of the
        derivative of :python:`x` are obtained via
        :python:`derivative_mat @ x`.
        """
        return self._derivative_mat

    @property
    def derivative2_mat(self) -> ndarray:
        """Matrix to create Fourier coefficients of the second derivative of
        some quantity.

        For some array of Fourier coefficients :python:`x`
        (excluding the sine component of the highest freequency), size
        :python:`(2*nfreq, ndof)`, the Fourier coefficients of the
        second derivative of :python:`x` are obtained via
        :python:`derivative2_mat @ x`.
        """
        return self._derivative2_mat

    @property
    def dt(self) -> float:
        """Time spacing [s]."""
        return self._time[1]

    @property
    def tf(self) -> float:
        """Final time (repeat period) [s]. Not included in
        :python:`time` vector.
        """
        return 1/self.f1

    @property
    def nt(self) -> int:
        """Number of timesteps."""
        return self.ncomponents

    @property
    def ncomponents(self) -> int:
        """Number of Fourier components (:python:`2*nfreq`) for each
        degree of freedom. Note that the sine component of the highest
        frequency (the 2-point wave) is excluded as this will always
        evaluate to zero.
        """
        return ncomponents(self.nfreq)

    @property
    def nstate_wec(self) -> int:
        """Length of the WEC dynamics state vector consisting of the
        Fourier coefficient of the position of each degree of freedom.
        """
        return self.ndof * self.ncomponents

    # other methods
    def decompose_state(self,
        state: ndarray
    ) -> tuple[ndarray, ndarray]:
        """Split the state vector into the WEC dynamics state and the
        optimization (control) state.

        Calls :py:func:`wecopttool.decompose_state` with the
        appropriate inputs for the WEC object.

        Examples
        --------
        >>> x_wec, x_opt = wec.decompose_state(x)

        Parameters
        ----------
        state
            Combined WEC and optimization states.

        Returns
        -------
        state_wec
            WEC state vector.
        state_opt
            Optimization (control) state.

        See Also
        --------
        decompose_state
        """
        return decompose_state(state, self.ndof, self.nfreq)

    def time_nsubsteps(self, nsubsteps: int) -> ndarray:
        """Create a time vector with finer discretization.

        Calls :py:func:`wecopttool.time` with the appropriate
        inputs for the WEC object.

        Parameters
        ----------
        nsubsteps
            Number of substeps between implied/default time steps.

        See Also
        --------
        time, WEC.time
        """
        return time(self.f1, self.nfreq, nsubsteps)

    def time_mat_nsubsteps(self, nsubsteps: int) -> ndarray:
        """Create a time matrix similar to
        :py:meth:`wecopttool.WEC.time_mat` but with finer
        time-domain discretization.

        Calls :py:func:`wecopttool.time_mat` with the appropriate
        inputs for the WEC object.

        Parameters
        ----------
        nsubsteps
            Number of substeps between implied/default time steps.

        See Also
        --------
        time_mat, WEC.time_mat, WEC.time_nsubsteps
        """
        return time_mat(self.f1, self.nfreq, nsubsteps)

    def vec_to_dofmat(self, vec: ndarray) -> ndarray:
        """Convert a vector to a matrix with one column per degree of
        freedom.

        Opposite of :py:meth:`wecopttool.WEC.dofmat_to_vec`.

        Calls :py:func:`wecopttool.vec_to_dofmat` with the
        appropriate inputs for the WEC object.

        Examples
        --------
        >>> x_wec, x_opt = wec.decompose_state(x)
        >>> x_wec_mat = wec.vec_to_dofmat(x_wec)

        Parameters
        ----------
        vec
            One-dimensional vector.

        See Also
        --------
        vec_to_dofmat, WEC.dofmat_to_vec
        """
        return vec_to_dofmat(vec, self.ndof)

    def dofmat_to_vec(self, mat: ndarray) -> ndarray:
        """Flatten a matrix to a vector.

        Opposite of :py:meth:`wecopttool.WEC.vec_to_dofmat`.

        Calls :py:func:`wecopttool.dofmat_to_vec` with the
        appropriate inputs for the WEC object.

        Parameters
        ----------
        mat
            Matrix with one column per degree of freedom.

        See Also
        --------
        dofmat_to_vec, WEC.vec_to_dofmat
        """
        return dofmat_to_vec(mat)

    def fd_to_td(self, fd: ndarray) -> ndarray:
        """Convert a frequency-domain array to time-domain.

        Opposite of :py:meth:`wecopttool.WEC.td_to_fd`.

        Calls :py:func:`wecopttool.fd_to_td` with the appropriate inputs
        for the WEC object.

        Parameters
        ----------
        fd
            Frequency-domain complex array with shape
            :python:`(WEC.nfreq+1, N)` for any :python:`N`.

        See Also
        --------
        fd_to_td, WEC.td_to_fd
        """
        return fd_to_td(fd, self.f1, self.nfreq, True)

    def td_to_fd(
        self,
        td: ndarray,
        fft: Optional[bool] = True,
        ) -> ndarray:
        """Convert a time-domain array to frequency-domain.

        Opposite of :py:meth:`wecopttool.WEC.fd_to_td`.

        Calls :py:func:`wecopttool.fd_to_td` with the appropriate
        inputs for the WEC object.

        Parameters
        ----------
        td
            Time-domain real array with shape
            :python:`(2*WEC.nfreq, N)` for any :python:`N`.
        fft
            Whether to use the real FFT.

        See Also
        --------
        td_to_fd, WEC.fd_to_td
        """
        return td_to_fd(td, fft, True)


def ncomponents(
    nfreq : int,
    zero_freq: Optional[bool] = True,
) -> int:
    """Number of Fourier components (:python:`2*nfreq`) for each
    DOF. The sine component of the highest frequency (the 2-point wave)
    is excluded as it will always evaluate to zero.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`X0` is excluded, and the number of components is reduced by 1.

    Parameters
    ----------
    nfreq
        Number of frequencies.
    zero_freq
        Whether to include the zero-frequency.
    """
    ncomp = 2*nfreq - 1
    if zero_freq:
        ncomp = ncomp + 1
    return ncomp


def frequency(
    f1: float,
    nfreq: int,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Construct equally spaced frequency array.

    The array includes :python:`0` and has length of :python:`nfreq+1`.
    :python:`f1` is fundamental frequency (1st harmonic).

    Returns the frequency array, e.g.,
    :python:`freqs = [0, f1, 2*f1, ..., nfreq*f1]`.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`0` is excluded, and the vector length is reduced by 1.

    Parameters
    ----------
    f1
        Fundamental frequency :python:`f1` [:math:`Hz`].
    nfreq
        Number of frequencies.
    zero_freq
        Whether to include the zero-frequency.
    """
    freq = np.arange(0, nfreq+1)*f1
    freq = freq[1:] if not zero_freq else freq
    return freq


def time(
    f1: float,
    nfreq: int,
    nsubsteps: Optional[int] = 1,
) -> ndarray:
    """Assemble the time vector with :python:`nsubsteps` subdivisions.

    Returns the 1D time vector, in seconds, starting at time
    :python:`0`, and not containing the end time :python:`tf=1/f1`.
    The time vector has length :python:`(2*nfreq)*nsubsteps`.
    The timestep length is :python:`dt = dt_default * 1/nsubsteps`,
    where :python:`dt_default=tf/(2*nfreq)`.

    Parameters
    ----------
    f1
        Fundamental frequency :python:`f1` [:math:`Hz`].
    nfreq
        Number of frequencies.
    nsubsteps
        Number of steps between the default (implied) time steps.
        A value of :python:`1` corresponds to the default step length.
    """
    if nsubsteps < 1:
        raise ValueError("'nsubsteps' must be 1 or greater")
    nsteps = nsubsteps * ncomponents(nfreq)
    return np.linspace(0, 1/f1, nsteps, endpoint=False)


def time_mat(
    f1: float,
    nfreq: int,
    nsubsteps: Optional[int] = 1,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Assemble the time matrix that converts the state to a
    time-series.

    For a state :math:`x` consisting of the mean (DC) component
    followed by the real and imaginary components of the Fourier
    coefficients (excluding the imaginary component of the 2-point wave) as
    :math:`x=[X0, Re(X1), Im(X1), ..., Re(Xn)]`,
    the response vector in the time-domain (:math:`x(t)`) is given as
    :math:`Mx`, where :math:`M` is the time matrix.

    The time matrix has size :python:`(nfreq*2, nfreq*2)`.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`X0` is excluded, and the matrix/vector length is reduced by 1.

    Parameters
    ---------
    f1
        Fundamental frequency :python:`f1` [:math:`Hz`].
    nfreq
        Number of frequencies.
    nsubsteps
        Number of steps between the default (implied) time steps.
        A value of :python:`1` corresponds to the default step length.
    zero_freq
        Whether the first frequency should be zero.
    """
    t = time(f1, nfreq, nsubsteps)
    omega = frequency(f1, nfreq) * 2*np.pi
    wt = np.outer(t, omega[1:])
    ncomp = ncomponents(nfreq)
    time_mat = np.empty((nsubsteps*ncomp, ncomp))
    time_mat[:, 0] = 1.0
    time_mat[:, 1::2] = np.cos(wt)
    time_mat[:, 2::2] = -np.sin(wt[:, :-1]) # remove 2pt wave sine component
    if not zero_freq:
        time_mat = time_mat[:, 1:]
    return time_mat


def derivative_mat(
    f1: float,
    nfreq: int,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Assemble the derivative matrix that converts the state vector of
    a response to the state vector of its derivative.

    For a state :math:`x` consisting of the mean (DC) component
    followed by the real and imaginary components of the Fourier
    coefficients (excluding the imaginary component of the 2-point wave) as
    :math:`x=[X0, Re(X1), Im(X1), ..., Re(Xn)]`,
    the state of its derivative is given as :math:`Dx`, where
    :math:`D` is the derivative matrix.

    The time matrix has size :python:`(nfreq*2, nfreq*2)`.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`X0` is excluded, and the matrix/vector length is reduced by 1.

    Parameters
    ---------
    f1
        Fundamental frequency :python:`f1` [:math:`Hz`].
    nfreq
        Number of frequencies.
    zero_freq
        Whether the first frequency should be zero.
    """
    def block(n): return np.array([[0, -1], [1, 0]]) * n*f1 * 2*np.pi
    blocks = [block(n+1) for n in range(nfreq)]
    if zero_freq:
        blocks = [0.0] + blocks
    deriv_mat = block_diag(*blocks)
    return deriv_mat[:-1, :-1] # remove 2pt wave sine component


def derivative2_mat(
    f1: float,
    nfreq: int,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Assemble the second derivative matrix that converts the state vector of
    a response to the state vector of its second derivative.

    For a state :math:`x` consisting of the mean (DC) component
    followed by the real and imaginary components of the Fourier
    coefficients (excluding the imaginary component of the 2-point wave) as
    :math:`x=[X0, Re(X1), Im(X1), ..., Re(Xn)]`,
    the state of its second derivative is given as :math:`(DD)x`, where
    :math:`DD` is the second derivative matrix.

    The time matrix has size :python:`(nfreq*2, nfreq*2)`.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`X0` is excluded, and the matrix/vector length is reduced by 1.

    Parameters
    ---------
    f1
        Fundamental frequency :python:`f1` [:math:`Hz`].
    nfreq
        Number of frequencies.
    zero_freq
        Whether the first frequency should be zero.
    """
    vals = [((n+1)*f1 * 2*np.pi)**2 for n in range(nfreq)]
    diagonal = np.repeat(-np.ones(nfreq) * vals, 2)[:-1] # remove 2pt wave sine
    if zero_freq:
        diagonal = np.concatenate(([0.0], diagonal))
    return np.diag(diagonal)


def mimo_transfer_mat(
    transfer_mat: ArrayLike,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Create a block matrix of the MIMO transfer function.

    The input is a complex transfer matrix that relates the complex
    Fourier representation of two variables.
    For example, it can be an impedance matrix or an RAO transfer
    matrix.
    The input complex impedance matrix has shape
    :python`(ndof, ndof, nfreq)`.

    Returns the 2D real matrix that transform the state representation
    of the input variable variable to the state representation of the
    output variable.
    Here, a state representation :python:`x` consists of the mean (DC)
    component followed by the real and imaginary components of the
    Fourier coefficients (excluding the imaginary component of the
    2-point wave) as
    :python:`x=[X0, Re(X1), Im(X1), ..., Re(Xn)]`.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`X0` is excluded, and the matrix/vector length is reduced by 1.

    Parameters
    ----------
    transfer_mat
        Complex transfer matrix.
    zero_freq
        Whether the first frequency should be zero.
    """
    ndof = transfer_mat.shape[0]
    assert transfer_mat.shape[1] == ndof
    elem = [[None]*ndof for _ in range(ndof)]
    def block(re, im): return np.array([[re, -im], [im, re]])
    for idof in range(ndof):
        for jdof in range(ndof):
            if zero_freq:
                Zp0 = transfer_mat[idof, jdof, 0]
                assert np.all(np.isreal(Zp0))
                Zp0 = np.real(Zp0)
                Zp = transfer_mat[idof, jdof, 1:]
            else:
                Zp0 = [0.0]
                Zp = transfer_mat[idof, jdof, :]
            re = np.real(Zp)
            im = np.imag(Zp)
            blocks = [block(ire, iim) for (ire, iim) in zip(re[:-1], im[:-1])]
            blocks = [Zp0] + blocks + [re[-1]]
            elem[idof][jdof] = block_diag(*blocks)
    return np.block(elem)


def vec_to_dofmat(vec: ArrayLike, ndof: int) -> ndarray:
    """Convert a vector back to a matrix with one column per DOF.

    Returns a matrix with :python:`ndof` columns.
    The number of rows is inferred from the size of the input vector.

    Opposite of :py:func:`wecopttool.dofmat_to_vec`.

    Parameters
    ----------
    vec
        1D array consisting of concatenated arrays of several DOFs, as
        :python:`vec = [vec_1, vec_2, ..., vec_ndof]`.
    ndof
        Number of degrees of freedom.

    See Also
    --------
    dofmat_to_vec,
    """
    return np.reshape(vec, (-1, ndof), order='F')


def dofmat_to_vec(mat: ArrayLike) -> ndarray:
    """Flatten a matrix that has one column per DOF.

    Returns a 1D vector.

    Opposite of :py:func:`wecopttool.vec_to_dofmat`.

    Parameters
    ----------
    mat
        Matrix to be flattened.

    See Also
    --------
    vec_to_dofmat,
    """
    return np.reshape(mat, -1, order='F')


def real_to_complex(
    fd: ArrayLike,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Convert from two real amplitudes to one complex amplitude per
    frequency.

    The input is a real 2D array with each column containing the real
    and imaginary components of the Fourier coefficients for some
    response, excluding the imaginary component of the highest frequency
    (2-point wave).
    The column length is :python:`2*nfreq`.
    The entries of a column representing a response :python:`x` are
    :python:`x=[X0, Re(X1), Im(X1), ..., Re(Xn)]`.

    Returns a complex 2D array with each column containing the complex
    Fourier coefficients.
    Columns are length :python:`nfreq+1`, and the first row corresponds
    to the real-valued zero-frequency (mean, DC) components.
    The entries of a column representing a response :python:`x` are
    :python:`x=[X0, X1, ..., Xn]`.

    If :python:`zero_freq = False`, the mean (DC) component :python:`X0`
    is excluded, and the column length is reduced by 1.

    Parameters
    ----------
    fd
        Array containing the real and imaginary components of the
        Fourier coefficients.
    zero_freq
        Whether the mean (DC) component is included.

    See Also
    --------
    complex_to_real,
    """
    fd= atleast_2d(fd)
    if zero_freq:
        assert fd.shape[0]%2==0
        mean = fd[0:1, :]
        fd = fd[1:, :]
    fdc = np.append(fd[0:-1:2, :] + 1j*fd[1::2, :],
                    [fd[-1, :]], axis=0)
    if zero_freq:
        fdc = np.concatenate((mean, fdc), axis=0)
    return fdc


def complex_to_real(
    fd: ArrayLike,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Convert from one complex amplitude to two real amplitudes per
    frequency.

    The input is a complex 2D array with each column containing the
    Fourier coefficients for some response.
    Columns are length :python:`nfreq+1`, and the first row corresponds
    to the real-valued zero-frequency (mean, DC) components.
    The entries of a column representing a response :python:`x` are
    :python:`x=[X0, X1, ..., Xn]`.

    Returns a real 2D array with each column containing the real and
    imaginary components of the Fourier coefficients. The imaginary component
    of the highest frequency (the 2-point wave) is excluded, as it will
    always evaluate to zero.
    The column length is :python:`2*nfreq`.
    The entries of a column representing a response :python:`x` are
    :python:`x=[X0, Re(X1), Im(X1), ..., Re(Xn)]`.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`X0` is excluded, and the vector length is reduced by 1.

    Parameters
    ----------
    fd
        Array containing the complex Fourier coefficients.
    zero_freq
        Whether the mean (DC) component is included.

    See Also
    --------
    real_to_complex,
    """
    fd = atleast_2d(fd)
    nfreq = fd.shape[0] - 1 if zero_freq else fd.shape[0]
    ndof = fd.shape[1]
    if zero_freq:
        assert np.all(np.isreal(fd[0, :]))
        a = np.real(fd[0:1, :])
        b = np.real(fd[1:-1, :])
        c = np.imag(fd[1:-1, :])
        d = np.real(fd[-1:, :])
    else:
        b = np.real(fd[:-1, :])
        c = np.imag(fd[:-1, :])
        d = np.real(fd[-1:, :])
    out = np.concatenate([np.transpose(b), np.transpose(c)])
    out = np.reshape(np.reshape(out, [-1], order='F'), [-1, ndof])
    if zero_freq:
        out = np.concatenate([a, out, d])
        assert out.shape == (2*nfreq, ndof)
    else:
        out = np.concatenate([out, d])
        assert out.shape == (2*nfreq-1, ndof)
    return out


def fd_to_td(
    fd: ArrayLike,
    f1: Optional[float] = None,
    nfreq: Optional[int] = None,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Convert a complex array of Fourier coefficients to a real array
    of time-domain responses.

    The input is a complex 2D array with each column containing the
    Fourier coefficients for some response.
    Columns are length :python:`nfreq+1`, and the first row corresponds
    to the real-valued zero-frequency (mean, DC) components.
    The entries of a column representing a response :python:`x` are
    :python:`x=[X0, X1, ..., Xn]`.

    Returns a real array with same number of columns and
    :python:`2*nfreq` rows, containing the time-domain response at
    times :python:`wecopttool.time(f1, nfreq, nsubsteps=1)`.
    The imaginary component of the highest frequency (the 2-point wave) is
    excluded, as it will always evaluate to zero.

    If both :python:`f1` and :python:`nfreq` are provided, it uses the
    time matrix :python:`wecopttool.time_mat(f1, nfreq, nsubsteps=1)`,
    else it uses the inverse real FFT (:py:func:`numpy.fft.irfft`).

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`X0` is excluded, and the matrix/vector length is reduced by 1.

    Opposite of :py:meth:`wecopttool.td_to_fd`.

    Parameters
    ----------
    fd
        Array containing the complex Fourier coefficients.
    f1
        Fundamental frequency :python:`f1` [:math:`Hz`].
    nfreq
        Number of frequencies.
    zero_freq
        Whether the mean (DC) component is included.

    Raises
    ------
    ValueError
        If only one of :python:`f1` or :python:`nfreq` is provided.
        Must provide both or neither.

    See Also
    --------
    td_to_fd, time, time_mat
    """
    fd = atleast_2d(fd)

    if zero_freq:
        msg = "The first row must be real when `zero_freq=True`."
        assert np.allclose(np.imag(fd[0, :]), 0), msg

    if (f1 is not None) and (nfreq is not None):
        tmat = time_mat(f1, nfreq, zero_freq=zero_freq)
        td = tmat @ complex_to_real(fd, zero_freq)
    elif (f1 is None) and (nfreq is None):
        n = 2*(fd.shape[0]-1)
        td = np.fft.irfft(fd/2, n=n, axis=0, norm='forward')
    else:
        raise ValueError(
            "Provide either both 'f1' and 'nfreq' or neither.")
    return td


def td_to_fd(
    td: ArrayLike,
    fft: Optional[bool] = True,
    zero_freq: Optional[bool] = True,
) -> ndarray:
    """Convert a real array of time-domain responses to a complex array
    of Fourier coefficients.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    :python:`X0` is excluded, and the matrix/vector length is reduced by 1.

    Opposite of :py:func:`wecopttool.fd_to_td`

    Parameters
    ----------
    td
        Real array of time-domains responses.
    fft
        Whether to use the real FFT.
    zero_freq
        Whether the mean (DC) component is returned.

    See Also
    --------
    fd_to_td
    """
    td= atleast_2d(td)
    n = td.shape[0]
    if fft:
        fd = np.fft.rfft(td*2, n=n, axis=0, norm='forward')
    else:
        fd = np.dot(dft(n, 'n')[:n//2+1, :], td*2)
    fd = np.concatenate((fd[:1, :]/2, fd[1:-1, :], fd[-1:, :]/2))
    if not zero_freq:
        fd = fd[1:, :]
    return fd


def read_netcdf(fpath: Union[str, Path]) -> Dataset:
    """Read a *NetCDF* file with possibly complex entries as a
    :py:class:`xarray.Dataset`.

    Can handle complex entries in the *NetCDF* by using
    :py:func:`capytaine.io.xarray.merge_complex_values`.

    Parameters
    ----------
    fpath
        Path to the *NetCDF* file.

    See Also
    --------
    write_netcdf,
    """
    with xr.open_dataset(fpath) as ds:
        ds.load()
    return cpy.io.xarray.merge_complex_values(ds)


def write_netcdf(fpath: Union[str, Path], data: Dataset) -> None:
    """Save an :py:class:`xarray.Dataset` with possibly complex entries as a
    *NetCDF* file.

    Can handle complex entries in the *NetCDF* by using
    :py:func:`capytaine.io.xarray.separate_complex_values`

    Parameters
    ----------
    fpath
        Name of file to save.
    data
        Dataset to save.

    See Also
    --------
    read_netcdf,
    """
    cpy.io.xarray.separate_complex_values(data).to_netcdf(fpath)


def check_linear_damping(
    hydro_data: Dataset,
    min_damping: Optional[float] = 1e-6,
    uniform_shift: Optional[bool] = True,
) -> Dataset:
    """Ensure that the linear hydrodynamics (friction + radiation
    damping) have positive damping.

    Shifts the :python:`friction` or :python:`radiation_damping` up
    if necessary. Returns the (possibly) updated Dataset with
    :python:`damping` :math:`>=` :python:`min_damping`.

    Parameters
    ----------
    hydro_data
        Linear hydrodynamic data.
    min_damping
        Minimum threshold for damping. Default is 1e-6.
    uniform_shift
        Boolean that determines whether the damping correction for each
        degree of freedom is frequency dependent or not. If :python:`True`,
        the damping correction is applied to :python:`friction` and shifts the
        damping for all frequencies. If :python:`False`, the damping correction
        is applied to :python:`radiation_damping` and only shifts the 
        damping for frequencies with negative damping values. Default is
        :python:`True`.
    """
    hydro_data_new = hydro_data.copy(deep=True)
    radiation = hydro_data_new['radiation_damping']
    friction = hydro_data_new['friction']
    ndof = len(hydro_data_new.influenced_dof)
    assert ndof == len(hydro_data.radiating_dof)
    for idof in range(ndof):
        iradiation = radiation.isel(radiating_dof=idof, influenced_dof=idof)
        ifriction = friction.isel(radiating_dof=idof, influenced_dof=idof)
        if uniform_shift:
            dmin = (iradiation+ifriction).min()
            if dmin <= 0.0 + min_damping:
                dof = hydro_data_new.influenced_dof.values[idof]
                delta = min_damping-dmin
                _log.warning(
                    f'Linear damping for DOF "{dof}" has negative or close ' +
                    'to zero terms. Shifting up via linear friction of ' +
                    f'{delta.values} N/(m/s).')
                hydro_data_new['friction'][idof, idof] = (ifriction + delta)
        else:
            new_damping = iradiation.where(
                iradiation+ifriction>min_damping, other=min_damping)
            dof = hydro_data_new.influenced_dof.values[idof]
            if (new_damping==min_damping).any():
                _log.warning(
                    f'Linear damping for DOF "{dof}" has negative or close to ' +
                    'zero terms. Shifting up damping terms to a minimum of ' +
                    f'{min_damping} N/(m/s)')
            hydro_data_new['radiation_damping'][:, idof, idof] = new_damping
    return hydro_data_new


def check_impedance(
    Zi: ArrayLike,
    min_damping: Optional[float] = 1e-6,
) -> DataArray:
    """Ensure that the real part of the impedance (resistive) is positive.

    Adds to real part of the impedance.
    Returns the (possibly) updated impedance with
    :math:`Re(Zi)>=` :python:`min_damping`.

    Parameters
    ----------
    Zi
        Linear hydrodynamic impedance.
    min_damping
        Minimum threshold for damping. Default is 1e-6.
    """
    Zi_diag = np.diagonal(Zi,axis1=0,axis2=1)
    Zi_shifted = Zi.copy()
    for dof in range(Zi_diag.shape[1]):
        dmin = np.min(np.real(Zi_diag[:, dof]))
        if dmin < min_damping:
            delta = min_damping - dmin
            Zi_shifted[dof,dof,:] = Zi_diag[:, dof] \
                + np.abs(delta)
            _log.warning(
                f'Real part of impedance for {dof} has negative or close to ' +
                f'zero terms. Shifting up by {delta:.2f}')
    return Zi_shifted


def force_from_rao_transfer_function(
    rao_transfer_mat: ArrayLike,
    zero_freq: Optional[bool] = True,
) -> TStateFunction:
    """Create a force function from its position transfer matrix.

    This is the position equivalent to the velocity-based
    :py:func:`wecopttool.force_from_impedance`.

    If :python:`zero_freq = False` (not default), the mean (DC) component
    of the transfer matrix (first row) is excluded.

    Parameters
    ----------
    rao_transfer_mat
        Complex position transfer matrix.
    zero_freq
        Whether the first frequency should be zero. Default is
        :python:`True`.

    See Also
    --------
    force_from_impedance,
    """
    def force(wec, x_wec, x_opt, waves):
        transfer_mat = mimo_transfer_mat(rao_transfer_mat, zero_freq)
        force_fd = wec.vec_to_dofmat(np.dot(transfer_mat, x_wec))
        return np.dot(wec.time_mat, force_fd)
    return force


def force_from_impedance(
    omega: ArrayLike,
    impedance: ArrayLike,
) -> TStateFunction:
    """Create a force function from its impedance.

    Parameters
    ----------
    omega
        Radial frequency vector.
    impedance
        Complex impedance matrix.

    See Also
    --------
    force_from_rao_transfer_function,
    """
    return force_from_rao_transfer_function(impedance*(1j*omega), False)


def force_from_waves(force_coeff: ArrayLike,
                     ) -> TStateFunction:
    """Create a force function from waves excitation coefficients.

    Parameters
    ----------
    force_coeff
        Complex excitation coefficients indexed by frequency and
        direction angle.
    """
    def force(wec, x_wec, x_opt, waves):
        force_fd = complex_to_real(wave_excitation(force_coeff, waves), False)
        return np.dot(wec.time_mat[:, 1:], force_fd)
    return force


def inertia(
    f1: float,
    nfreq: int,
    inertia_matrix: ArrayLike,
) -> TStateFunction:
    """Create the inertia "force" from the inertia matrix.

    Parameters
    ----------
    f1
        Fundamental frequency :python:`f1` [:math:`Hz`].
    nfreq
        Number of frequencies.
    inertia_matrix
        Inertia matrix.
    """
    omega = np.reshape(frequency(f1, nfreq, False)*2*np.pi, [1,1,-1])
    inertia_matrix = np.expand_dims(inertia_matrix, -1)
    rao_transfer_function = -1*omega**2*inertia_matrix + 0j
    inertia_fun = force_from_rao_transfer_function(
        rao_transfer_function, False)
    return inertia_fun


def standard_forces(hydro_data: Dataset) -> TForceDict:
    """Create functions for linear hydrodynamic forces.

    Returns a dictionary with the standard linear forces:
    radiation, hydrostatic, friction, Froude—Krylov, and diffraction.
    The functions are type :python:`StateFunction` (see Type Aliases in
    API Documentation).

    Parameters
    ----------
    hydro_data
        Linear hydrodynamic data.
    """
    hydro_data = hydro_data.transpose(
         "omega", "wave_direction", "radiating_dof", "influenced_dof")

    # intrinsic impedance
    w = hydro_data['omega']
    A = hydro_data['added_mass']
    B = hydro_data['radiation_damping']
    K = hydro_data['hydrostatic_stiffness']
    Bf = hydro_data['friction']

    rao_transfer_functions = dict()
    rao_transfer_functions['radiation'] = (1j*w*B + -1*w**2*A, False)
    rao_transfer_functions['friction'] = (1j*w*Bf, False)

    # include zero_freq in hydrostatics
    hs = ((K + 0j).expand_dims({"omega": B.omega}))
    tmp = hs.isel(omega=0).copy(deep=True)
    tmp['omega'] = tmp['omega'] * 0
    hs = xr.concat([tmp, hs], dim='omega') #, data_vars='minimal')
    rao_transfer_functions['hydrostatics'] = (hs, True)

    linear_force_functions = dict()
    for name, (value, zero_freq) in rao_transfer_functions.items():
        value = value.transpose("radiating_dof", "influenced_dof", "omega")
        value = -1*value  # RHS of equation: ma = Σf
        linear_force_functions[name] = (
            force_from_rao_transfer_function(value, zero_freq))

    # wave excitation
    excitation_coefficients = {
        'Froude_Krylov': hydro_data['Froude_Krylov_force'],
        'diffraction': hydro_data['diffraction_force']
    }

    for name, value in excitation_coefficients.items():
        linear_force_functions[name] = force_from_waves(value)

    return linear_force_functions


def run_bem(
    fb: cpy.FloatingBody,
    freq: Iterable[float] = [np.infty],
    wave_dirs: Iterable[float] = [0],
    rho: float = _default_parameters['rho'],
    g: float = _default_parameters['g'],
    depth: float = _default_parameters['depth'],
    write_info: Optional[Mapping[str, bool]] = None,
    njobs: int = 1,
) -> Dataset:
    """Run Capytaine for a range of frequencies and wave directions.

    This simplifies running *Capytaine* and ensures the output are in
    the correct convention (see
    :py:func:`wecopttool.change_bem_convention`).

    It creates the *test matrix*, calls
    :py:meth:`capytaine.bodies.bodies.FloatingBody.keep_immersed_part`,
    calls :py:meth:`capytaine.bem.solver.BEMSolver.fill_dataset`,
    and changes the sign convention using
    :py:func:`wecopttool.change_bem_convention`.

    Parameters
    ----------
    fb
        The WEC as a Capytaine floating body (mesh + DOFs).
    freq
        List of frequencies [:math:`Hz`] to evaluate BEM at.
    wave_dirs
        List of wave directions [degrees] to evaluate BEM at.
    rho
        Water density in :math:`kg/m^3`.
    g
        Gravitational acceleration in :math:`m/s^2`.
    depth
        Water depth in :math:`m`.
    write_info
        Which additional information to write.
        Options are:
        :python:`['hydrostatics', 'mesh', 'wavelength', 'wavenumber']`.
        See :py:func:`capytaine.io.xarray.assemble_dataset` for more
        details.
    njobs
        Number of jobs to run in parallel.
        See :py:meth:`capytaine.bem.solver.BEMSolver.fill_dataset`

    See Also
    --------
    change_bem_convention,
    """
    if wave_dirs is not None:
        wave_dirs = np.atleast_1d(degrees_to_radians(wave_dirs))
    solver = cpy.BEMSolver()
    test_matrix = Dataset(coords={
        'rho': [rho],
        'water_depth': [depth],
        'omega': [ifreq*2*np.pi for ifreq in freq],
        'wave_direction': wave_dirs,
        'radiating_dof': list(fb.dofs.keys()),
        'g': [g],
    })
    if wave_dirs is None:
        # radiation only problem, no diffraction or excitation
        test_matrix = test_matrix.drop_vars('wave_direction')
    if write_info is None:
        write_info = {'hydrostatics': False,
                      'mesh': False,
                      'wavelength': False,
                      'wavenumber': False,
                     }
    wec_im = fb.copy(name=f"{fb.name}_immersed").keep_immersed_part()
    bem_data = solver.fill_dataset(
        test_matrix, wec_im, n_jobs=njobs, **write_info)
    return change_bem_convention(bem_data)


def change_bem_convention(bem_data: Dataset) -> Dataset:
    """Change the convention from :math:`-iωt` to :math:`+iωt`.

    Change the linear hydrodynamic coefficients from the Capytaine
    convention (:math:`x(t)=Xe^{-iωt}`), where :math:`X` is the
    frequency-domain response, to the more standard convention
    used in WecOptTool (:math:`x(t)=Xe^{+iωt}`).

    NOTE: This might change in Capytaine in the future.

    Parameters
    ----------
    bem_data
        Linear hydrodynamic coefficients for the WEC.
    """
    bem_data['Froude_Krylov_force'] = np.conjugate(
        bem_data['Froude_Krylov_force'])
    bem_data['diffraction_force'] = np.conjugate(bem_data['diffraction_force'])
    return bem_data


def linear_hydrodynamics(
    bem_data: Dataset,
    inertia_matrix: Optional[ArrayLike] = None,
    hydrostatic_stiffness: Optional[ArrayLike] = None,
    friction: Optional[ArrayLike] = None
) -> Dataset:
    """Add rigid body inertia_matrix, hydrostatic stiffness, and linear
    friction to BEM data.

    Returns the Dataset with the additional information added.

    Parameters
    ----------
    bem_data
        Linear hydrodynamic coefficients obtained using the boundary
        element method (BEM) code Capytaine, with sign convention
        corrected.
    inertia_matrix
        Inertia matrix of size :python:`(ndof, ndof)`.
        :python:`None` if included in :python:`bem_data`.
    hydrostatic_stiffness
        Linear hydrostatic restoring coefficient of size
        :python:`(nodf, ndof)`.
        :python:`None` if included in :python:`bem_data`.
    friction
        Linear friction, in addition to radiation damping, of size
        :python:`(nodf, ndof)`.
        :python:`None` if included in :python:`bem_data` or to set to zero.

    Raises
    ------
    ValueError
        If either :python:`inertia_matrix` or
        :python:`hydrostatic_stiffness` are :python:`None` and is not
        included in :python:`bem_data`.
    ValueError
        If any of :python:`inertia_matrix`,
        :python:`hydrostatic_stiffness`, or :python:`friction` are both
        provided and included in :python:`bem_data` but have different
        values.
    """
    vars = {'inertia_matrix': inertia_matrix, 'friction': friction,
            'hydrostatic_stiffness': hydrostatic_stiffness}

    dims = ['radiating_dof', 'influenced_dof']

    hydro_data = bem_data.copy(deep=True)

    for name, data in vars.items():
        org = name in hydro_data.variables.keys()
        new = data is not None
        if new and org:
            if not np.allclose(data, hydro_data.variables[name]):
                raise ValueError(
                    f'BEM data already has variable "{name}" ' +
                    'with diferent values')
            else :
                _log.warning(
                    f'Variable "{name}" is already in BEM data ' +
                    'with same value.')
        elif (not new) and (not org):
            if name=='friction':
                ndof = len(hydro_data["influenced_dof"])
                hydro_data[name] = (dims, np.zeros([ndof, ndof]))
            else:
                raise ValueError(
                    f'Variable "{name}" is not in BEM data and ' +
                    'was not provided.')
        elif new:
            data = atleast_2d(data)
            hydro_data[name] = (dims, data)

    return hydro_data


def wave_excitation(exc_coeff: Dataset, waves: Dataset) -> ndarray:
    """Calculate the complex, frequency-domain, excitation force due to
    waves.

    The resulting force is indexed only by frequency and not direction
    angle.
    The input :python:`waves` frequencies must be same as
    :python:`exc_coeff`, but the directions can be a subset.

    Parameters
    ----------
    exc_coeff
        Complex excitation coefficients indexed by frequency and
        direction angle.
    waves
        Complex frequency-domain wave elevation.

    Raises
    ------
    ValueError
        If the frequency vectors of :python:`exc_coeff` and
        :python:`waves` are different.
    ValueError
        If any of the directions in :python:`waves` is not in
        :python:`exc_coeff`.
    """
    omega_w = waves['omega'].values
    omega_e = exc_coeff['omega'].values
    dir_w = waves['wave_direction'].values
    dir_e = exc_coeff['wave_direction'].values
    exc_coeff = exc_coeff.transpose(
        'omega', 'wave_direction', 'influenced_dof').values

    wave_elev_fd = np.expand_dims(waves.values, -1)

    if not np.allclose(omega_w, omega_e):
        raise ValueError(f"Wave and excitation frequencies do not match. WW: {omega_w}, EE: {omega_e}")

    subset, sub_ind = subset_close(dir_w, dir_e)

    if not subset:
        raise ValueError(
            "Some wave directions are not in excitation coefficients " +
            f"\n Wave direction(s): {(np.rad2deg(dir_w))} (deg)" +
            f"\n BEM direction(s): {np.rad2deg(dir_e)} (deg).")

    return np.sum(wave_elev_fd*exc_coeff[:, sub_ind, :], axis=1)


def hydrodynamic_impedance(hydro_data: Dataset) -> Dataset:
    """Calculate hydrodynamic intrinsic impedance.

    Parameters
    ----------
    hydro_data
        Dataset with linear hydrodynamic coefficients produced by
        :py:func:`wecopttool.linear_hydrodynamics`.
    """

    Zi = (hydro_data['inertia_matrix'] \
        + hydro_data['added_mass'])*1j*hydro_data['omega'] \
            + hydro_data['radiation_damping'] + hydro_data['friction'] \
                + hydro_data['hydrostatic_stiffness']/1j/hydro_data['omega']
    return Zi


def atleast_2d(array: ArrayLike) -> ndarray:
    """Ensure an array is at least 2D, otherwise add trailing dimensions
    to make it 2D.

    This differs from :py:func:`numpy.atleast_2d` in that the additional
    dimensions are appended at the end rather than at the begining.
    This might be an option in :py:func:`numpy.atleast_2d` in the
    future, see
    `NumPy #12336 <https://github.com/numpy/numpy/issues/12336>`_.

    Parameters
    ----------
    array
        Input array.
    """
    array = np.atleast_1d(array)
    return np.expand_dims(array, -1) if len(array.shape)==1 else array


def degrees_to_radians(
    degrees: FloatOrArray,
    sort: Optional[bool] = True,
) -> Union[float, ndarray]:
    """Convert a 1D array of angles in degrees to radians in the range
    :math:`[-π, π)` and optionally sort them.

    Parameters
    ----------
    degrees
        1D array of angles in degrees.
    sort
        Whether to sort the angles from smallest to largest in
        :math:`[-π, π)`.
    """
    radians = np.asarray(np.remainder(np.deg2rad(degrees), 2*np.pi))
    radians[radians > np.pi] -= 2*np.pi
    if radians.size == 1:
        radians = radians.item()
    elif sort:
        radians = np.sort(radians)
    return radians


def subset_close(
    set_a: FloatOrArray,
    set_b: FloatOrArray,
    rtol: float = 1.e-5,
    atol: float = 1.e-8,
    equal_nan: bool = False,
) -> tuple[bool, list]:
    """Check if the first set :python:`set_a` is contained, to some
    tolerance, in the second set :python:`set_b`.

    Parameters
    ----------
    set_a
        First array which is tested for being subset.
    set_b
        Second array which is tested for containing :python:`set_a`.
    rtol
        The relative tolerance parameter. Passed to
        :py:func:`numpy.isclose`.
    atol
        The absolute tolerance parameter. Passed to
        :py:func:`numpy.isclose`.
    equal_nan
        Whether to compare NaNs as equal. Passed to
        :py:func:`numpy.isclose`.

    Returns
    -------
    subset
        Whether the first array is a subset of the second array.
    ind
        List with integer indices where the first array's elements are
        located inside the second array.
        Only contains values if :python:`subset==True`.

    Raises
    ------
    ValueError
        If either of the two arrays contains repeated elements.
    """
    set_a = np.atleast_1d(set_a)
    set_b = np.atleast_1d(set_b)

    if len(np.unique(set_a.round(decimals = 6))) != len(set_a):
        raise ValueError("Elements in set_a not unique")
    if len(np.unique(set_b.round(decimals = 6))) != len(set_b):
        raise ValueError("Elements in set_b not unique")

    ind = []
    tmp_result = [False for _ in range(len(set_a))]
    for subset_element in set_a:
        for set_element in set_b:
            if np.isclose(subset_element, set_element, rtol, atol, equal_nan):
                tmp_set_ind = np.where(
                    np.isclose(set_element, set_b , rtol, atol, equal_nan))
                tmp_subset_ind = np.where(
                    np.isclose(subset_element, set_a , rtol, atol,
                               equal_nan))
                ind.append( int(tmp_set_ind[0]) )
                tmp_result[ int(tmp_subset_ind[0]) ] = True
    subset = all(tmp_result)
    ind = ind if subset else []
    return subset, ind


def scale_dofs(scale_list: Iterable[float], ncomponents: int) -> ndarray:
    """Create a scaling vector based on a different scale for each DOF.

    Returns a 1D array of length :python:`NDOF*ncomponents` where the
    number of DOFs (:python:`NDOF`) is the length of
    :python:`scale_list`.
    The first :python:`ncomponents` entries have the value of the first
    scale :python:`scale_list[0]`, the next :python:`ncomponents`
    entries have the value of the second scale :python:`scale_list[1]`,
    and so on.


    Parameters
    ----------
    scale_list
        Scale for each DOF.
    ncomponents
        Number of elements in the state vector for each DOF.
    """
    ndof = len(scale_list)
    scale = []
    for dof in range(ndof):
        scale += [scale_list[dof]] * ncomponents
    return np.array(scale)


def decompose_state(
    state: ndarray,
    ndof: int,
    nfreq: int,
) -> tuple[ndarray, ndarray]:
    """Split the state vector into the WEC dynamics state and the
    optimization (control) state.

    The WEC dynamics state consists of the Fourier coefficients of
    the position of each degree of freedom.
    The optimization state depends on the chosen control states for
    the problem.

    Parameters
    ----------
    state
        Combined WEC and optimization states.
    ndof
        Number of degrees of freedom for the WEC dynamics.
    nfreq
        Number of frequencies.

    Returns
    -------
    state_wec
        WEC state vector.
    state_opt
        Optimization (control) state.
    """
    nstate_wec = ndof * ncomponents(nfreq)
    return state[:nstate_wec], state[nstate_wec:]


def frequency_parameters(
    freqs: ArrayLike,
    zero_freq: bool = True,
) -> tuple[float, int]:
    """Return the fundamental frequency and the number of frequencies
    in a frequency array.

    This function can be used as a check for inputs to other functions
    since it raises an error if the frequency vector does not have
    the correct format :python:`freqs = [0, f1, 2*f1, ..., nfreq*f1]`
    (or :python:`freqs = [f1, 2*f1, ..., nfreq*f1]` if
    :python:`zero_freq = False`).

    Parameters
    ----------
    freqs
        The frequency array, starting at zero and having equal spacing.
    zero_freq
        Whether the first frequency should be zero.

    Returns
    -------
    f1
        Fundamental frequency :python:`f1` [:math:`Hz`]
    nfreq
        Number of frequencies (not including zero frequency),
        i.e., :python:`freqs = [0, f1, 2*f1, ..., nfreq*f1]`.

    Raises
    ------
    ValueError
        If the frequency vector is not evenly spaced.
    ValueError
        If the zero-frequency was expected but not included or not
        expected but included.
    """
    if np.isclose(freqs[0], 0.0):
        if zero_freq:
            freqs0 = freqs[:]
        else:
            raise ValueError('Zero frequency was included.')
    else:
        if zero_freq:
            raise ValueError(
                'Frequency array must start with the zero frequency.')
        else:
            freqs0 = np.concatenate([[0.0,], freqs])

    f1 = freqs0[1]
    nfreq = len(freqs0) - 1
    f_check = np.arange(0, f1*(nfreq+0.5), f1)
    if not np.allclose(f_check, freqs0):
        raise ValueError("Frequency array 'omega' must be evenly spaced by" +
                         "the fundamental frequency " +
                         "(i.e., 'omega = [0, f1, 2*f1, ..., nfreq*f1]')")
    return f1, nfreq


def time_results(fd: DataArray, time: DataArray) -> ndarray:
    """Create a :py:class:`xarray.DataArray` of time-domain results from
    :py:class:`xarray.DataArray` of frequency-domain results.

    Parameters
    ----------
    fd
        Frequency domain response.
    time
        Time array.
    """
    out = np.zeros((*fd.isel(omega=0).shape, len(time)))
    for w, mag in zip(fd.omega, fd):
        out = out + \
            np.real(mag)*np.cos(w*time) - np.imag(mag)*np.sin(w*time)
    return out



"""Geometry and parameters for some example devices (WEC).
"""


from __future__ import annotations


__all__ = [
    "WaveBot",
]

from typing import Optional

import matplotlib.pyplot as plt
from matplotlib.pyplot import Axes, Figure
from meshio._mesh import Mesh
import gmsh
import pygmsh


class WaveBot:
    """Class representing the Sandia WaveBot.

    See, e.g.,

        - https://doi.org/10.3390/en10040472
        - https://doi.org/10.2172/1330189

    """

    def __init__(self,
                 r1: float = 0.88,
                 r2: float = 0.35,
                 h1: float = 0.17,
                 h2: float = 0.37,
                 scale_factor: float = 1,
                 freeboard: float = 0.01,
                 ) -> None:
        """Create a WaveBot with specific dimensions.

        Parameters
        ----------
        r1
            Outer-most radius (of cylindrical section) [m].
        r2
            Inner-most radius (of the conic frustum) [m].
        h1
            Height of the cylindrical section [m].
        h2
            Height of the conic frustum section [m].
        scale_factor
            Scale factor to linearly scale hull dimensions [ ].
        freeboard
            Freeboard above free surface (will be removed later for BEM
            calculations) [m]. The draft of the cylindrical section is
            :python:`h1-freeboard`.
        """
        self.r1 = r1 * scale_factor
        self.r2 = r2 * scale_factor
        self.h1 = h1 * scale_factor
        self.h2 = h2 * scale_factor
        self.freeboard = freeboard * scale_factor
        self.scale_factor = scale_factor
        self.gear_ratio = 12.47

    def mesh(self, mesh_size_factor: Optional[float] = 0.1) -> Mesh:
        """Generate surface mesh of hull.

        Parameters
        ----------
        mesh_size_factor
            Control for the mesh size. Smaller values give a finer mesh.
        """
        with pygmsh.occ.Geometry() as geom:
            gmsh.option.setNumber('Mesh.MeshSizeFactor', mesh_size_factor)
            cyl = geom.add_cylinder([0, 0, 0],
                                    [0, 0, -self.h1],
                                    self.r1)
            cone = geom.add_cone([0, 0, -self.h1],
                                 [0, 0, -self.h2],
                                 self.r1, self.r2)
            geom.translate(cyl, [0, 0, self.freeboard])
            geom.translate(cone, [0, 0, self.freeboard])
            geom.boolean_union([cyl, cone])
            mesh = geom.generate_mesh()

        return mesh

    def plot_cross_section(self,
                           show: bool = False,
                           ax: Optional[Axes] = None,
                           **kwargs,
                           ) -> tuple[Figure, Axes]:
        """
        Plot hull cross-section.

        Parameters
        ----------
        show
            Whether to show the figure.
        ax
            Existing axes. The default is None.
            If None, new axes will be created.
        **kwargs
            Passed to :func:`matplotlib.pyplot.plot`.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None

        y = [-1*(self.h1 - self.freeboard + self.h2),
             -1*(self.h1 - self.freeboard + self.h2),
             -1*(self.h1 - self.freeboard),
             0,
             self.freeboard,
             self.freeboard]
        x = [0,
             self.r2,
             self.r1,
             self.r1,
             self.r1,
             0]
        ax.plot(x, y,
                marker='.',
                **kwargs)

        ax.set_xlim(left=0)
        ax.axhline(0, color='grey', linestyle='--')
        ax.set_xlabel('Radius [m]')
        ax.set_ylabel('Height [m]')
        ax.axis('equal')

        if show:
            plt.show()

        return fig, ax


class AquaHarmonics:
    """Class representing the AquaHarmonics WEC.

    See https://aquaharmonics.com/wec_vis/.
    """

    def __init__(self,
                 T1: float = 1.5,
                 T2: float = 0.355,
                 T3: float = 7.25,
                 r1: float = 1.085,
                 r2: float = 0.405,
                 r3: float = 0.355,
                 scale_factor: float = 1,
                 ofst: float = 0.1,
                 ) -> None:
        """
        Create the AquaHarmonics WEC with specific dimensions.

        Parameters
        ----------
        T1 : float, optional
            Draft of large cylinder [m].
        T2 : float, optional
            Height of conic frustum [m].
        T3 : float, optional
            Height of tube [m].
        r1 : float, optional
            Radius of larger cylinder [m].
        r2 : float, optional
            Outer radius of tube [m].
        r3 : float, optional
            Inner radius of tube [m].
        scale_factor
            Scale factor to linearly scale hull dimensions [ ].
        ofst : float, optional
            Offset for clipping at waterplane [m].
        """
        self.T1 = T1 * scale_factor
        self.T2 = T2 * scale_factor
        self.T3 = T3 * scale_factor
        self.r1 = r1 * scale_factor
        self.r2 = r2 * scale_factor
        self.r3 = r3 * scale_factor
        self.ofst = ofst * scale_factor
        self.scale_factor = scale_factor

    def mesh(self, mesh_size_factor: float = 0.25) -> Mesh:
        """
        Mesh of AquaHarmonics hull.

        Returns
        -------
        mesh
            Surface mesh of hull.

        """
        with pygmsh.occ.Geometry() as geom:
            gmsh.option.setNumber('Mesh.MeshSizeFactor', mesh_size_factor)
            cyl1 = geom.add_cylinder([0, 0, 0],
                                     [0, 0, -self.T1],
                                     self.r1)
            cone = geom.add_cone([0, 0, -self.T1],
                                 [0, 0, -self.T2],
                                 self.r1,
                                 self.r2)
            cylout = geom.add_cylinder([0, 0, -1*(self.T1+self.T2)],
                                       [0, 0, -self.T3],
                                       self.r2)
            cylin = geom.add_cylinder([0, 0, -1*(self.T1+self.T2)],
                                      [0, 0, -self.T3],
                                      self.r3)
            cyl2 = geom.boolean_difference(cylout, cylin)[0]
            wecGeom = geom.boolean_union(entities=[cyl1, cone, cyl2],
                                         delete_first=True)[0]

            geom.translate(wecGeom, [0, 0, self.ofst])
            mesh = geom.generate_mesh()

        return mesh

    def plot_cross_section(self,
                           show: bool = False,
                           ax: Optional[Axes] = None,
                           **kwargs,
                           ) -> tuple[Figure, Axes]:
        """
        Plot hull cross-section.

        Parameters
        ----------
        show
            Whether to show the figure.
        ax
            Existing axes. The default is None.
            If None, new axes will be created.
        **kwargs
            Passed to :func:`matplotlib.pyplot.plot`.
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None

        ax.plot(
            [0, self.r3, self.r3, self.r2, self.r2, self.r1, self.r1],
            [-1*(self.T1+self.T2),
             -1*(self.T1+self.T2),
             -1*(self.T1+self.T2+self.T3),
             -1*(self.T1+self.T2+self.T3),
             -1*(self.T1+self.T2),
             -1*(self.T1),
             self.ofst],
            marker='.'
        )

        ax.set_xlim(left=0)
        ax.axhline(0, color='grey', linestyle='--')
        ax.axvline(0, color='k', linestyle='--')
        ax.set_xlabel('Radius [m]')
        ax.set_ylabel('Height [m]')
        ax.axis('equal')

        if show:
            plt.show()

        return fig, ax
